# coding=utf-8
# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import functools
import glob
import importlib
import os
import re
import sys
import tarfile
import zipfile
from collections import defaultdict
from contextlib import closing
from email.message import Message
from email.parser import Parser
from io import StringIO
from textwrap import dedent

from pex import pex_warnings
from pex.common import open_zip, pluralize
from pex.compatibility import to_unicode
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.third_party.packaging.markers import Marker
from pex.third_party.packaging.requirements import InvalidRequirement
from pex.third_party.packaging.requirements import Requirement as PackagingRequirement
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        DefaultDict,
        Dict,
        FrozenSet,
        Iterable,
        Iterator,
        List,
        Optional,
        Text,
        Tuple,
        Union,
    )

    import attr  # vendor:skip

    from pex.pep_440 import ParsedVersion
else:
    from pex.third_party import attr


class MetadataError(Exception):
    """Indicates an error reading distribution metadata."""


class UnrecognizedDistributionFormat(MetadataError):
    """Indicates a distribution file is not of any recognized format."""


class AmbiguousDistributionError(MetadataError):
    """Indicates multiple distributions were detected at a given location but one was expected."""


class MetadataNotFoundError(MetadataError):
    """Indicates an expected metadata file could not be found for a given distribution."""


_PKG_INFO_BY_DIST_LOCATION = {}  # type: Dict[Text, Optional[Message]]


def _strip_sdist_path(sdist_path):
    # type: (Text) -> Optional[str]
    if not sdist_path.endswith((".sdist", ".tar.gz", ".tgz", ".tar.bz2", ".tbz2", ".zip")):
        return None

    sdist_basename = os.path.basename(sdist_path)
    filename, _ = os.path.splitext(sdist_basename)
    if filename.endswith(".tar"):
        filename, _ = os.path.splitext(filename)
    # All PEP paths lead here for the definition of a valid project name which limits things to
    # ascii; so this str(...) is Python 2.7 safe: https://peps.python.org/pep-0508/#names
    # The version part of the basename is similarly restricted by:
    #   https://peps.python.org/pep-0440/#summary-of-changes-to-pep-440
    return str(filename)


def _parse_message(message):
    # type: (bytes) -> Message
    return cast(Message, Parser().parse(StringIO(to_unicode(message))))


def _parse_sdist_package_info(sdist_path):
    # type: (Text) -> Optional[Message]
    sdist_filename = _strip_sdist_path(sdist_path)
    if sdist_filename is None:
        return None

    pkg_info_path = os.path.join(sdist_filename, "PKG-INFO")

    if zipfile.is_zipfile(sdist_path):
        with open_zip(sdist_path) as zip:
            try:
                return _parse_message(zip.read(pkg_info_path))
            except KeyError as e:
                pex_warnings.warn(
                    "Source distribution {} did not have the expected metadata file {}: {}".format(
                        sdist_path, pkg_info_path, e
                    )
                )
                return None

    if tarfile.is_tarfile(sdist_path):
        with tarfile.open(sdist_path) as tf:
            try:
                pkg_info = tf.extractfile(pkg_info_path)
                if pkg_info is None:
                    # N.B.: `extractfile` returns None for directories and special files.
                    return None
                with closing(pkg_info) as fp:
                    return _parse_message(fp.read())
            except KeyError as e:
                pex_warnings.warn(
                    "Source distribution {} did not have the expected metadata file {}: {}".format(
                        sdist_path, pkg_info_path, e
                    )
                )
                return None

    return None


@attr.s(frozen=True)
class DistMetadataFile(object):
    project_name = attr.ib()  # type: ProjectName
    version = attr.ib()  # type: Version
    path = attr.ib()  # type: str


def find_dist_info_files(
    filename,  # type: Text
    listing,  # type: Iterable[str]
):
    # type: (...) -> Iterator[DistMetadataFile]

    # N.B. We know the captured project_name and version will not contain `-` even though PEP-503
    # allows for them in project names and PEP-440 allows for them in versions in some
    # circumstances. This is since we're limiting ourselves to the products of installs by our
    # vendored versions of wheel and pip which turn `-` into `_` as explained in `ProjectName` and
    # `Version` docs.
    dist_info_metadata_pattern = "^{}$".format(
        os.path.join(r"(?P<project_name>.+)-(?P<version>.+)\.dist-info", re.escape(filename))
    )
    wheel_metadata_re = re.compile(dist_info_metadata_pattern)
    for item in listing:
        match = wheel_metadata_re.match(item)
        if match:
            yield DistMetadataFile(
                project_name=ProjectName(match.group("project_name")),
                version=Version(match.group("version")),
                path=item,
            )


def find_dist_info_file(
    project_name,  # type: Union[Text, ProjectName]
    filename,  # type: Text
    listing,  # type: Iterable[str]
    version=None,  # type: Optional[Union[Text, Version]]
):
    # type: (...) -> Optional[str]

    normalized_project_name = (
        project_name if isinstance(project_name, ProjectName) else ProjectName(project_name)
    )

    if isinstance(version, Version):
        normalized_version = version
    elif isinstance(version, str):
        normalized_version = Version(version)
    else:
        normalized_version = None

    for metadata_file in find_dist_info_files(filename, listing):
        if normalized_project_name == metadata_file.project_name:
            if normalized_version and normalized_version != metadata_file.version:
                continue
            return metadata_file.path
    return None


def _parse_wheel_package_info(wheel_path):
    # type: (Text) -> Optional[Message]
    if not wheel_path.endswith(".whl") or not zipfile.is_zipfile(wheel_path):
        return None
    project_name, version, _ = os.path.basename(wheel_path).split("-", 2)
    with open_zip(wheel_path) as whl:
        metadata_file = find_dist_info_file(
            project_name=project_name,
            version=version,
            filename="METADATA",
            listing=whl.namelist(),
        )
        if not metadata_file:
            return None
        with whl.open(metadata_file) as fp:
            return _parse_message(fp.read())


def _parse_installed_distribution_info(location):
    # type: (Text) -> Optional[Message]

    if not os.path.isdir(location):
        return None

    dist_info_dirs = glob.glob(os.path.join(location, "*.dist-info"))
    if not dist_info_dirs:
        return None

    if len(dist_info_dirs) > 1:
        raise AmbiguousDistributionError(
            "Found more than one distribution at {location}:\n{dist_info_dirs}".format(
                location=location,
                dist_info_dirs="\n".join(
                    os.path.relpath(dist_info_dir, location) for dist_info_dir in dist_info_dirs
                ),
            )
        )

    metadata_file = os.path.join(dist_info_dirs[0], "METADATA")
    if not os.path.exists(metadata_file):
        return None

    with open(metadata_file, "rb") as fp:
        return _parse_message(fp.read())


def _parse_pkg_info(location):
    # type: (Text) -> Optional[Message]
    if location not in _PKG_INFO_BY_DIST_LOCATION:
        pkg_info = _parse_wheel_package_info(location)
        if not pkg_info:
            pkg_info = _parse_sdist_package_info(location)
        if not pkg_info:
            pkg_info = _parse_installed_distribution_info(location)
        _PKG_INFO_BY_DIST_LOCATION[location] = pkg_info
    return _PKG_INFO_BY_DIST_LOCATION[location]


@attr.s(frozen=True)
class ProjectNameAndVersion(object):
    @classmethod
    def from_parsed_pkg_info(cls, source, pkg_info):
        # type: (str, Message) -> ProjectNameAndVersion
        project_name = pkg_info.get("Name", None)
        version = pkg_info.get("Version", None)
        if project_name is None or version is None:
            raise MetadataError(
                "The 'Name' and 'Version' fields are not both present in package metadata for "
                "{source}:\n{fields}".format(
                    source=source,
                    fields="\n".join("{}: {}".format(k, v) for k, v in pkg_info.items()),
                )
            )
        return cls(project_name=pkg_info["Name"], version=pkg_info["Version"])

    @classmethod
    def from_filename(cls, path):
        # type: (Text) -> ProjectNameAndVersion
        # Handle wheels:
        #
        # The wheel filename convention is specified here:
        #   https://www.python.org/dev/peps/pep-0427/#file-name-convention.
        if path.endswith(".whl"):
            project_name, version, _ = os.path.basename(path).split("-", 2)
            return cls(project_name=project_name, version=version)

        # Handle sdists:
        #
        # The sdist name format has no accepted specification yet, but there is a proposal here:
        #   https://www.python.org/dev/peps/pep-0625/#specification.
        #
        # We do the best we can to support the current landscape. A version number can technically
        # contain a dash though, even under the standards, in un-normalized form:
        #   https://www.python.org/dev/peps/pep-0440/#pre-release-separators.
        # For those cases this logic will produce incorrect results and it does not seem there is
        # much we can do since both project names and versions can contain both alphanumeric
        # characters and dashes.
        fname = _strip_sdist_path(path)
        if fname is not None:
            components = fname.rsplit("-", 1)
            if len(components) == 2:
                project_name, version = components
                return cls(project_name=project_name, version=version)

        raise UnrecognizedDistributionFormat(
            "The distribution at path {!r} does not have a file name matching known sdist or wheel "
            "file name formats.".format(path)
        )

    project_name = attr.ib()  # type: Text
    version = attr.ib()  # type: Text

    @property
    def canonicalized_project_name(self):
        # type: () -> ProjectName
        return ProjectName(self.project_name)

    @property
    def canonicalized_version(self):
        # type: () -> Version
        return Version(self.version)


def project_name_and_version(
    location,  # type: Union[Text, Distribution, Message]
    fallback_to_filename=True,  # type: bool
):
    # type: (...) -> Optional[ProjectNameAndVersion]
    """Extracts name and version metadata from dist.

    :param location: A distribution to extract project name and version metadata from.
    :return: The project name and version.
    :raise: MetadataError if dist has invalid metadata.
    """
    if isinstance(location, Distribution):
        return ProjectNameAndVersion(project_name=location.project_name, version=location.version)

    pkg_info = location if isinstance(location, Message) else _parse_pkg_info(location)
    if pkg_info is not None:
        if isinstance(location, str):
            source = location
        else:
            source = "<parsed message>"
        return ProjectNameAndVersion.from_parsed_pkg_info(source=source, pkg_info=pkg_info)
    if fallback_to_filename and not isinstance(location, (Distribution, Message)):
        return ProjectNameAndVersion.from_filename(location)
    return None


def requires_python(location):
    # type: (Union[Text, Distribution, Message]) -> Optional[SpecifierSet]
    """Examines dist for `Python-Requires` metadata and returns version constraints if any.

    See: https://www.python.org/dev/peps/pep-0345/#requires-python

    :param location: A distribution to check for `Python-Requires` metadata.
    :return: The required python version specifiers.
    """
    if isinstance(location, Distribution):
        return location.metadata.requires_python

    pkg_info = location if isinstance(location, Message) else _parse_pkg_info(location)
    if pkg_info is None:
        return None

    python_requirement = pkg_info.get("Requires-Python", None)
    if python_requirement is None:
        return None
    return SpecifierSet(python_requirement)


def requires_dists(location):
    # type: (Union[Text, Distribution, Message]) -> Iterator[Requirement]
    """Examines dist for and returns any declared requirements.

    Looks for `Requires-Dist` metadata.

    The older `Requires` metadata is intentionally ignored, although we do log a warning if it is
    found to draw attention to this ~work-around and the associated issue in case any new data
    comes in.

    See:
    + https://www.python.org/dev/peps/pep-0345/#requires-dist-multiple-use
    + https://www.python.org/dev/peps/pep-0314/#requires-multiple-use

    :param location: A distribution to check for requirement metadata.
    :return: All requirements found.
    """
    if isinstance(location, Distribution):
        for requirement in location.metadata.requires_dists:
            yield requirement
        return

    pkg_info = location if isinstance(location, Message) else _parse_pkg_info(location)
    if pkg_info is None:
        return

    for requires_dist in pkg_info.get_all("Requires-Dist", ()):
        yield Requirement.parse(requires_dist)

    legacy_requires = pkg_info.get_all("Requires", [])  # type: List[str]
    if legacy_requires:
        name_and_version = project_name_and_version(location)
        project_name = name_and_version.project_name if name_and_version else location
        pex_warnings.warn(
            dedent(
                """\
                Ignoring {count} `Requires` {field} in {dist} metadata:
                {requires}

                You may have issues using the '{project_name}' distribution as a result.
                More information on this workaround can be found here:
                  https://github.com/pantsbuild/pex/issues/1201#issuecomment-791715585
                """
            ).format(
                dist=location,
                project_name=project_name,
                count=len(legacy_requires),
                field=pluralize(legacy_requires, "field"),
                requires=os.linesep.join(
                    "{index}.) Requires: {req}".format(index=index, req=req)
                    for index, req in enumerate(legacy_requires, start=1)
                ),
            )
        )


class RequirementParseError(Exception):
    """Indicates and invalid requirement string.

    See PEP-508: https://www.python.org/dev/peps/pep-0508
    """


@attr.s(frozen=True)
class Requirement(object):
    @classmethod
    def parse(cls, requirement):
        # type: (Text) -> Requirement
        try:
            return cls.from_packaging_requirement(PackagingRequirement(requirement))
        except InvalidRequirement as e:
            raise RequirementParseError(str(e))

    @classmethod
    def from_packaging_requirement(cls, requirement):
        # type: (PackagingRequirement) -> Requirement
        return cls(
            name=requirement.name,
            url=requirement.url,
            extras=frozenset(requirement.extras),
            specifier=requirement.specifier,
            marker=requirement.marker,
        )

    name = attr.ib(eq=False)  # type: str
    url = attr.ib(default=None)  # type: Optional[str]
    extras = attr.ib(default=frozenset())  # type: FrozenSet[str]
    specifier = attr.ib(factory=SpecifierSet)  # type: SpecifierSet
    marker = attr.ib(default=None, eq=str)  # type: Optional[Marker]

    project_name = attr.ib(init=False, repr=False)  # type: ProjectName
    _str = attr.ib(init=False, eq=False, repr=False)  # type: str

    def __attrs_post_init__(self):
        object.__setattr__(self, "project_name", ProjectName(self.name))

        parts = [self.name]
        if self.extras:
            parts.append("[{extras}]".format(extras=",".join(sorted(self.extras))))
        if self.specifier:
            parts.append(str(self.specifier))
        if self.url:
            parts.append("@ {url}".format(url=self.url))
            if self.marker:
                parts.append(" ")
        if self.marker:
            parts.append("; {marker}".format(marker=self.marker))
        object.__setattr__(self, "_str", "".join(parts))

    @property
    def key(self):
        # type: () -> str
        return self.project_name.normalized

    def __contains__(self, item):
        # type: (Union[str, Version, Distribution, ProjectNameAndVersion]) -> bool

        # We emulate pkg_resources.Requirement.__contains__ pre-release behavior here since the
        # codebase expects it.
        return self.contains(item, prereleases=True)

    def contains(
        self,
        item,  # type: Union[str, Version, Distribution, ProjectNameAndVersion]
        prereleases=None,  # type: Optional[bool]
    ):
        # type: (...) -> bool
        if isinstance(item, ProjectNameAndVersion):
            if item.canonicalized_project_name != self.project_name:
                return False
            version = item.canonicalized_version.parsed_version  # type: Union[ParsedVersion, str]
        elif isinstance(item, Distribution):
            if item.key != self.key:
                return False
            version = item.metadata.version.parsed_version
        elif isinstance(item, Version):
            version = item.parsed_version
        else:
            version = item

        # We know SpecifierSet.contains returns bool on inspection of its code. The fact we import
        # via the pex.third_party mechanism makes the type opaque to MyPy. We also know it accepts
        # either a "parsed_version" or a str and take advantage of this to save re-parsing version
        # strings we've already parsed.
        return cast(bool, self.specifier.contains(version, prereleases=prereleases))

    def __str__(self):
        # type: () -> str
        return self._str


# N.B.: DistributionMetadata can have an expensive hash when a distribution has many requirements;
# so we cache the hash. See: https://github.com/pantsbuild/pex/issues/1928
@attr.s(frozen=True, cache_hash=True)
class DistMetadata(object):
    @classmethod
    def load(cls, location):
        # type: (Union[Text, Message]) -> DistMetadata

        project_name_and_ver = project_name_and_version(location)
        if not project_name_and_ver:
            raise MetadataError(
                "Failed to determine project name and version for distribution at "
                "{location}.".format(location=location)
            )
        return cls(
            project_name=ProjectName(project_name_and_ver.project_name),
            version=Version(project_name_and_ver.version),
            requires_dists=tuple(requires_dists(location)),
            requires_python=requires_python(location),
        )

    project_name = attr.ib()  # type: ProjectName
    version = attr.ib()  # type: Version
    requires_dists = attr.ib(default=())  # type: Tuple[Requirement, ...]
    requires_python = attr.ib(default=SpecifierSet())  # type: Optional[SpecifierSet]


def _realpath(path):
    # type: (str) -> str
    return os.path.realpath(path)


@attr.s(frozen=True)
class Distribution(object):
    @staticmethod
    def _read_metadata_lines(metadata_path):
        # type: (str) -> Iterator[str]
        with open(os.path.join(metadata_path)) as fp:
            for line in fp:
                # This is pkg_resources.IMetadataProvider.get_metadata_lines behavior, which our
                # code expects.
                normalized = line.strip()
                if normalized and not normalized.startswith("#"):
                    yield normalized

    @classmethod
    def parse_entry_map(cls, entry_points_metadata_path):
        # type: (str) -> Dict[str, Dict[str, EntryPoint]]

        # This file format is defined here:
        #   https://packaging.python.org/en/latest/specifications/entry-points/#file-format

        entry_map = defaultdict(dict)  # type: DefaultDict[str, Dict[str, EntryPoint]]
        group = None  # type: Optional[str]
        for index, line in enumerate(cls._read_metadata_lines(entry_points_metadata_path), start=1):
            if line.startswith("[") and line.endswith("]"):
                group = line[1:-1]
            elif not group:
                raise ValueError(
                    "Failed to parse entry_points.txt, encountered an entry point with no "
                    "group on line {index}: {line}".format(index=index, line=line)
                )
            else:
                entry_point = EntryPoint.parse(line)
                entry_map[group][entry_point.name] = entry_point
        return entry_map

    @classmethod
    def load(cls, location):
        # type: (str) -> Distribution
        return cls(location=location, metadata=DistMetadata.load(location))

    # N.B.: Resolving the distribution location through any symlinks is pkg_resources behavior,
    # which our code expects.
    location = attr.ib(converter=_realpath)  # type: str

    metadata = attr.ib()  # type: DistMetadata
    _metadata_files_cache = attr.ib(
        factory=dict, init=False, eq=False, repr=False
    )  # type: Dict[str, str]

    @property
    def key(self):
        # type: () -> str
        return self.metadata.project_name.normalized

    @property
    def project_name(self):
        # type: () -> str
        return self.metadata.project_name.raw

    @property
    def version(self):
        # type: () -> str
        return self.metadata.version.raw

    def as_requirement(self):
        # type: () -> Requirement
        return Requirement(
            name=self.project_name,
            specifier=SpecifierSet("=={version}".format(version=self.version)),
        )

    def requires(self):
        # type: () -> Tuple[Requirement, ...]
        return self.metadata.requires_dists

    def _get_metadata_file(self, name):
        # type: (str) -> Optional[str]
        normalized_name = os.path.normpath(name)
        if os.path.isabs(normalized_name):
            raise ValueError(
                "The metadata file name must be a relative path under the .dist-info/ directory. "
                "Given: {name}".format(name=name)
            )

        metadata_file = self._metadata_files_cache.get(normalized_name)
        if metadata_file is None:
            metadata_file = find_dist_info_file(
                project_name=self.metadata.project_name,
                version=self.version,
                filename=normalized_name,
                listing=[
                    os.path.relpath(path, self.location)
                    for path in glob.glob(
                        os.path.join(
                            self.location, "*.dist-info/{name}".format(name=normalized_name)
                        )
                    )
                ],
            )
            # N.B.: We store the falsey "" as the sentinel that we've searched already and the
            # metadata file did not exist.
            self._metadata_files_cache[normalized_name] = metadata_file or ""
        return metadata_file or None

    def has_metadata(self, name):
        # type: (str) -> bool
        return self._get_metadata_file(name) is not None

    def get_metadata_lines(self, name):
        # type: (str) -> Iterator[str]
        relative_path = self._get_metadata_file(name)
        if relative_path is None:
            raise MetadataNotFoundError(
                "The metadata file {name} is not present for {project_name} {version} at "
                "{location}".format(
                    name=name,
                    project_name=self.project_name,
                    version=self.version,
                    location=self.location,
                )
            )
        for line in self._read_metadata_lines(os.path.join(self.location, relative_path)):
            yield line

    def get_entry_map(self):
        # type: () -> Dict[str, Dict[str, EntryPoint]]
        entry_points_metadata_relpath = self._get_metadata_file("entry_points.txt")
        if entry_points_metadata_relpath is None:
            return defaultdict(dict)
        return self.parse_entry_map(os.path.join(self.location, entry_points_metadata_relpath))

    def __str__(self):
        # type: () -> str
        return "{project_name} {version}".format(
            project_name=self.project_name, version=self.version
        )


@attr.s(frozen=True)
class EntryPoint(object):
    @classmethod
    def parse(cls, spec):
        # type: (str) -> EntryPoint

        # This file format is defined here:
        #   https://packaging.python.org/en/latest/specifications/entry-points/#file-format

        components = spec.split("=")
        if len(components) != 2:
            raise ValueError("Invalid entry point specification: {spec}.".format(spec=spec))

        name, value = components
        module, sep, attrs = value.strip().partition(":")
        if sep and not attrs:
            raise ValueError("Invalid entry point specification: {spec}.".format(spec=spec))

        entry_point_name = name.strip()
        if sep:
            return CallableEntryPoint(
                name=entry_point_name, module=module, attrs=tuple(attrs.split("."))
            )

        return cls(name=entry_point_name, module=module)

    name = attr.ib()  # type: str
    module = attr.ib()  # type: str

    def __str__(self):
        # type: () -> str
        return self.module


@attr.s(frozen=True)
class CallableEntryPoint(EntryPoint):
    _attrs = attr.ib()  # type: Tuple[str, ...]

    @_attrs.validator
    def _validate_attrs(self, _, value):
        if not value:
            raise ValueError("A callable entry point must select a callable item from the module.")

    def resolve(self):
        # type: () -> Callable[[], Any]
        module = importlib.import_module(self.module)
        try:
            return cast("Callable[[], Any]", functools.reduce(getattr, self._attrs, module))
        except AttributeError as e:
            raise ImportError(
                "Could not resolve {attrs} in {module}: {err}".format(
                    attrs=".".join(self._attrs), module=module, err=e
                )
            )

    def __str__(self):
        # type: () -> str
        return "{module}:{attrs}".format(module=self.module, attrs=".".join(self._attrs))


def find_distribution(
    project_name,  # type: Union[str, ProjectName]
    search_path=None,  # type: Optional[Iterable[str]]
):
    # type: (...) -> Optional[Distribution]

    for location in search_path or sys.path:
        if not os.path.isdir(location):
            continue

        metadata_file = find_dist_info_file(
            project_name=str(project_name),
            filename="METADATA",
            listing=[
                os.path.relpath(path, location)
                for path in glob.glob(os.path.join(location, "*.dist-info/METADATA"))
            ],
        )
        if not metadata_file:
            continue

        metadata_path = os.path.join(location, metadata_file)
        with open(metadata_path, "rb") as fp:
            pkg_info = _parse_message(fp.read())
            dist = Distribution(location=location, metadata=DistMetadata.load(pkg_info))
            if dist.metadata.project_name == (
                project_name if isinstance(project_name, ProjectName) else ProjectName(project_name)
            ):
                return dist

    return None


def find_distributions(search_path=None):
    # type: (Optional[Iterable[str]]) -> Iterator[Distribution]

    seen = set()
    for location in search_path or sys.path:
        if not os.path.isdir(location):
            continue
        for metadata_file in find_dist_info_files(
            filename="METADATA",
            listing=[
                os.path.relpath(path, location)
                for path in glob.glob(os.path.join(location, "*.dist-info/METADATA"))
            ],
        ):
            metadata_path = os.path.realpath(os.path.join(location, metadata_file.path))
            if metadata_path in seen:
                continue
            seen.add(metadata_path)
            with open(metadata_path, "rb") as fp:
                pkg_info = _parse_message(fp.read())
                yield Distribution(location=location, metadata=DistMetadata.load(pkg_info))



# Copyright 2016 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from contextlib import contextmanager
from textwrap import dedent

from pex.common import open_zip, safe_mkdtemp, temporary_dir
from pex.testing import WheelBuilder, make_project, pex_project_dir, temporary_content
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Dict, Iterable, Iterator, List, Optional, Union


BDIST_PEX_VENV = None  # type: Optional[Virtualenv]


def bdist_pex_venv():
    # type: () -> Virtualenv
    # In order to run the bdist_pex distutils command we need:
    # 1. setuptools on the sys.path since the test projects use and test setuptools.setup and its
    #    additional features above and beyond distutils.core.setup like entry points declaration.
    # 2. Pex on the sys.path so its distutils command module(s) can be found.
    # 3. An indication to distutils of where to look for Pex distutils commands.
    #
    # We take care of 1 and 2 here and 3 is taken care of by passing --command-packages to
    # distutils.

    global BDIST_PEX_VENV
    if BDIST_PEX_VENV is None:
        venv_dir = safe_mkdtemp()
        venv = Virtualenv.create(venv_dir)
        pip = venv.install_pip()
        # N.B.: The setuptools version is not important, but there is a break introduced by pip
        # 22.1; so we must pin that low.
        subprocess.check_call(args=[pip, "install", "-U", "pip<22.1"])
        subprocess.check_call(args=[pip, "install", pex_project_dir(), "setuptools==43.0.0"])
        BDIST_PEX_VENV = venv
    return BDIST_PEX_VENV


@contextmanager
def bdist_pex(project_dir, bdist_args=None):
    # type: (str, Optional[Iterable[str]]) -> Iterator[List[str]]
    with temporary_dir() as dist_dir:
        cmd = [
            "setup.py",
            "--command-packages",
            "pex.commands",
            "bdist_pex",
            "--bdist-dir={}".format(dist_dir),
        ]
        if bdist_args:
            cmd.extend(bdist_args)

        venv = bdist_pex_venv()
        _, process = venv.interpreter.open_process(args=cmd, cwd=project_dir)
        process.wait()
        yield [os.path.join(dist_dir, dir_entry) for dir_entry in os.listdir(dist_dir)]


def assert_entry_points(entry_points, bdist_args=None):
    # type: (Union[str, Dict[str, List[str]]], Optional[Iterable[str]]) -> Iterator[str]
    with make_project(name="my_app", entry_points=entry_points) as project_dir:
        with bdist_pex(project_dir, bdist_args) as apps_pex:
            for app_pex in apps_pex:
                process = subprocess.Popen([app_pex], stdout=subprocess.PIPE)
                stdout, _ = process.communicate()
                assert "{pex_root}" not in os.listdir(project_dir)
                assert 0 == process.returncode
                assert stdout == b"hello world!\n"
                yield os.path.basename(app_pex)


def assert_pex_args_shebang(shebang):
    # type: (str) -> None
    with make_project() as project_dir:
        pex_args = '--pex-args=--python-shebang="{}"'.format(shebang)
        with bdist_pex(project_dir, bdist_args=[pex_args]) as (my_app_pex,):
            with open(my_app_pex, "rb") as fp:
                assert fp.readline().decode().rstrip() == shebang


def test_entry_points_dict():
    # type: () -> None
    (_,) = assert_entry_points({"console_scripts": ["my_app = my_app.my_module:do_something"]})


def test_entry_points_ini_string():
    # type: () -> None
    (_,) = assert_entry_points(
        dedent(
            """
            [console_scripts]
            my_app=my_app.my_module:do_something
            """
        )
    )


def test_bdist_all_single_entry_point_dict():
    # type: () -> None
    assert {"first_app"} == set(
        assert_entry_points(
            {"console_scripts": ["first_app = my_app.my_module:do_something"]}, ["--bdist-all"]
        )
    )


def test_bdist_all_two_entry_points_dict():
    # type: () -> None
    assert {"first_app", "second_app"} == set(
        assert_entry_points(
            {
                "console_scripts": [
                    "first_app = my_app.my_module:do_something",
                    "second_app = my_app.my_module:do_something",
                ]
            },
            ["--bdist-all"],
        )
    )


def test_bdist_all_single_entry_point_ini_string():
    # type: () -> None
    (my_app,) = assert_entry_points(
        dedent(
            """
            [console_scripts]
            my_app=my_app.my_module:do_something
            """
        ),
        ["--bdist-all"],
    )
    assert "my_app" == my_app


def test_bdist_all_two_entry_points_ini_string():
    # type: () -> None
    assert {"first_app", "second_app"} == set(
        assert_entry_points(
            dedent(
                """
            [console_scripts]
            first_app=my_app.my_module:do_something
            second_app=my_app.my_module:do_something
            """
            ),
            ["--bdist-all"],
        )
    )


def test_pex_args_shebang_with_spaces():
    # type: () -> None
    assert_pex_args_shebang("#!/usr/bin/env python")


def test_pex_args_shebang_without_spaces():
    # type: () -> None
    assert_pex_args_shebang("#!/usr/bin/python")


def test_unwriteable_contents():
    # type: () -> None
    my_app_setup_py = dedent(
        """
        from setuptools import setup
        
        setup(
            name='my_app',
            version='0.0.0',
            zip_safe=True,
            packages=['my_app'],
            include_package_data=True,
            package_data={'my_app': ['unwriteable.so']},
        )
        """
    )

    UNWRITEABLE_PERMS = 0o400
    with temporary_content(
        {
            "setup.py": my_app_setup_py,
            "my_app/__init__.py": "",
            "my_app/unwriteable.so": "so contents",
        },
        perms=UNWRITEABLE_PERMS,
    ) as my_app_project_dir:
        my_app_whl = WheelBuilder(my_app_project_dir).bdist()

        with make_project(name="uses_my_app", install_reqs=["my_app"]) as uses_my_app_project_dir:
            pex_args = "--pex-args=--disable-cache --pip-version=vendored --no-pypi -f {}".format(
                os.path.dirname(my_app_whl)
            )
            with bdist_pex(uses_my_app_project_dir, bdist_args=[pex_args]) as (uses_my_app_pex,):
                with open_zip(uses_my_app_pex) as zf:
                    unwriteable_sos = [
                        path for path in zf.namelist() if path.endswith("my_app/unwriteable.so")
                    ]
                    assert 1 == len(unwriteable_sos)
                    unwriteable_so = unwriteable_sos.pop()
                    zf.extract(unwriteable_so, path=uses_my_app_project_dir)
                    extract_dest = os.path.join(uses_my_app_project_dir, unwriteable_so)
                    with open(extract_dest) as fp:
                        assert "so contents" == fp.read()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os.path
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.inherit_path import InheritPath
from pex.orderedset import OrderedSet
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Any, List


DUMP_SYS_PATH_CODE = dedent(
    """\
    import json
    import sys


    json.dump(sys.path, sys.stdout)
    """
)


def create_sys_path_dump_pex(
    tmpdir,  # type: Any
    *additional_args  # type: str
):
    # type: (...) -> str

    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write(DUMP_SYS_PATH_CODE)

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["--exe", exe, "-o", pex] + list(additional_args)).assert_success()
    return pex


def execute_sys_path_dump_pex(
    pex,  # type: str
    *additional_args,  # type: str
    **additional_env  # type: Any
):
    # type: (...) -> OrderedSet[str]

    return OrderedSet(
        os.path.realpath(entry)
        for entry in cast(
            "List[str]",
            json.loads(
                subprocess.check_output(
                    args=[pex] + list(additional_args), env=make_env(**additional_env)
                )
            ),
        )
    )


def read_additional_sys_path(
    pex,  # type: str
    *additional_args,  # type: str
    **additional_env  # type: Any
):
    # type: (...) -> List[str]

    isolated_sys_path = execute_sys_path_dump_pex(pex)
    return list(
        execute_sys_path_dump_pex(pex, *additional_args, **additional_env) - isolated_sys_path
    )


def assert_inherited(
    tmpdir,  # type: Any
    pex,  # type: str
    *expected_inherited_sys_path_entries,  # type: str
    **additional_env  # type: Any
):
    other_exe = os.path.join(str(tmpdir), "other_exe.py")
    with safe_open(other_exe, "w") as fp:
        fp.write(DUMP_SYS_PATH_CODE)

    subprocess_proof = os.path.join(str(tmpdir), "proof")
    assert not os.path.exists(subprocess_proof)

    code = dedent(
        """\
        import subprocess
        import sys


        open({subprocess_proof!r}, "w").close()
        sys.exit(subprocess.call([sys.executable, {other_exe!r}]))
        """
    ).format(subprocess_proof=subprocess_proof, other_exe=other_exe)
    actual = read_additional_sys_path(pex, "-c", code, PEX_INTERPRETER=1, **additional_env)
    assert os.path.exists(subprocess_proof)

    previous = -1
    for expected_inherited_sys_path_entry in expected_inherited_sys_path_entries:
        assert expected_inherited_sys_path_entry in actual
        current_index = actual.index(expected_inherited_sys_path_entry)
        assert current_index > previous
        previous = current_index


@pytest.mark.parametrize(
    "execution_mode_args",
    [
        pytest.param([], id="ZIPAPP"),
        pytest.param(["--venv"], id="VENV"),
    ],
)
def test_pex_run_extra_sys_path(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = create_sys_path_dump_pex(tmpdir, *execution_mode_args)

    foo = os.path.join(str(tmpdir), "foo")
    assert [foo] == read_additional_sys_path(pex, PEX_EXTRA_SYS_PATH=foo)

    bar = os.path.join(str(tmpdir), "bar")
    assert_inherited(tmpdir, pex, bar, PEX_EXTRA_SYS_PATH=bar)


@pytest.mark.parametrize(
    "inherit_path",
    [pytest.param(inherit_path, id=str(inherit_path)) for inherit_path in InheritPath.values()],
)
def test_pex_run_inherit_path_and_extra_sys_path(
    tmpdir,  # type: Any
    inherit_path,  # type: InheritPath.Value
):
    # type: (...) -> None

    pex = create_sys_path_dump_pex(
        tmpdir, "--inherit-path={inherit_path}".format(inherit_path=inherit_path)
    )

    esp = os.path.join(str(tmpdir), "esp")
    pp1 = os.path.join(str(tmpdir), "pp1")
    pp2 = os.path.join(str(tmpdir), "pp2")

    expected = [esp] if inherit_path is InheritPath.FALSE else [pp1, pp2, esp]
    assert_inherited(
        tmpdir, pex, *expected, PEX_EXTRA_SYS_PATH=esp, PYTHONPATH=os.pathsep.join((pp1, pp2))
    )



"""Run the EasyInstall command"""

if __name__ == '__main__':
    if "__PEX_UNVENDORED__" in __import__("os").environ:
      from setuptools.command.easy_install import main  # vendor:skip
    else:
      from pex.third_party.setuptools.command.easy_install import main

    main()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import hashlib
import os.path

import pytest

from pex.pep_503 import ProjectName
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.downloads import ArtifactDownloader
from pex.resolve.locked_resolve import Artifact, FileArtifact
from pex.resolve.resolved_requirement import Fingerprint, PartialArtifact
from pex.resolve.resolver_configuration import PipConfiguration
from pex.testing import IS_LINUX
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


def file_artifact(
    url,  # type: str
    sha256,  # type: str
):
    # type: (...) -> FileArtifact
    artifact = Artifact.from_url(
        url=url, fingerprint=Fingerprint(algorithm="sha256", hash=sha256), verified=True
    )
    assert isinstance(artifact, FileArtifact)
    return artifact


LINUX_ARTIFACT = file_artifact(
    url=(
        "https://files.pythonhosted.org/packages/6d/c6/"
        "6a4e46802e8690d50ba6a56c7f79ac283e703fcfa0fdae8e41909c8cef1f/"
        "psutil-5.9.1-cp310-cp310-"
        "manylinux_2_12_x86_64"
        ".manylinux2010_x86_64"
        ".manylinux_2_17_x86_64"
        ".manylinux2014_x86_64.whl"
    ),
    sha256="29a442e25fab1f4d05e2655bb1b8ab6887981838d22effa2396d584b740194de",
)

MAC_ARTIFACT = file_artifact(
    url=(
        "https://files.pythonhosted.org/packages/d1/16/"
        "6239e76ab5d990dc7866bc22a80585f73421588d63b42884d607f5f815e2/"
        "psutil-5.9.1-cp310-cp310-macosx_10_9_x86_64.whl"
    ),
    sha256="c7be9d7f5b0d206f0bbc3794b8e16fb7dbc53ec9e40bbe8787c6f2d38efcf6c9",
)


@pytest.fixture
def downloader():
    # type: () -> ArtifactDownloader
    return ArtifactDownloader(ConfiguredResolver(PipConfiguration()))


def test_issue_1849_download_foreign_artifact(
    tmpdir,  # type: str
    downloader,  # type: ArtifactDownloader
):
    # type: (...) -> None

    foreign_artifact = MAC_ARTIFACT if IS_LINUX else LINUX_ARTIFACT

    dest_dir = os.path.join(str(tmpdir), "dest_dir")
    assert foreign_artifact.filename == downloader.download(
        foreign_artifact, dest_dir=dest_dir, digest=hashlib.sha256()
    )


def test_issue_1849_fingerprint_foreign_artifact(
    tmpdir,  # type: str
    downloader,  # type: ArtifactDownloader
):
    # type: (...) -> None

    expected_artifacts = [LINUX_ARTIFACT, MAC_ARTIFACT]
    assert expected_artifacts == list(
        downloader.fingerprint([PartialArtifact(artifact.url) for artifact in expected_artifacts])
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import re

import pytest

from pex.pep_503 import ProjectName
from pex.resolve.pep_691.api import Client
from pex.resolve.pep_691.model import Endpoint, Project
from pex.resolve.resolved_requirement import Fingerprint
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Type


def request(
    url,  # type: str
    content_type,  # type: str
):
    # type: (...) -> Project
    return Client().request(Endpoint(url=url, content_type=content_type))


def assert_error(
    url,  # type: str
    content_type,  # type: str
    expected_exception_type,  # type: Type[Exception]
    prefix,  # type: str
):
    # type: (...) -> None

    with pytest.raises(expected_exception_type, match="^{}.*".format(re.escape(prefix))):
        request(url, content_type)


def test_invalid_content_type():
    # type: () -> None

    assert_error(
        url="https://example.org",
        content_type="text/html",
        expected_exception_type=ValueError,
        prefix=(
            "Asked to request project metadata from https://example.org for text/html but only the "
            "following API content types are accepted:"
        ),
    )


def test_invalid_json():
    # type: () -> None

    assert_error(
        url="https://example.org",
        content_type="application/vnd.pypi.simple.v1+json",
        expected_exception_type=Client.Error,
        prefix=(
            "PEP-691 API request to https://example.org for application/vnd.pypi.simple.v1+json "
            "returned invalid JSON: "
        ),
    )


def test_http_error():
    # type: () -> None

    assert_error(
        url="https://pypi.org/simple/p527-DNE",
        content_type="application/vnd.pypi.simple.v1+json",
        expected_exception_type=Client.Error,
        prefix=(
            "PEP-691 API request to https://pypi.org/simple/p527-DNE for "
            "application/vnd.pypi.simple.v1+json failed: HTTP Error 404: Not Found"
        ),
    )


@pytest.mark.parametrize(
    "content_type", [pytest.param(content_type, id=content_type) for content_type in Client.ACCEPT]
)
def test_valid(content_type):
    # type: (str) -> None

    p537 = request(url="https://pypi.org/simple/p537", content_type=content_type)
    assert ProjectName("p537") == p537.name

    files_by_filename = {file.filename: file for file in p537.files}
    assert (
        Fingerprint(
            algorithm="sha256",
            hash="c1300324000522387b1f0c474db53b081308f78868030e53a2df6f162d3feb86",
        )
        == files_by_filename["p537-1.0.0-cp36-cp36m-macosx_10_13_x86_64.whl"].select_fingerprint()
    )



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
import logging
import os.path
from argparse import ArgumentParser, _ActionsContainer

from pex import pex_warnings
from pex.cli.command import BuildTimeCommand
from pex.commands.command import JsonMixin, OutputMixin
from pex.common import DETERMINISTIC_DATETIME, is_script, open_zip, pluralize
from pex.dist_metadata import Distribution
from pex.enum import Enum
from pex.executor import Executor
from pex.pex import PEX
from pex.pex_info import PexInfo
from pex.resolve import configured_resolve, requirement_options, resolver_options, target_options
from pex.resolve.resolver_configuration import (
    LockRepositoryConfiguration,
    PexRepositoryConfiguration,
)
from pex.result import Error, Ok, Result, try_
from pex.targets import LocalInterpreter, Target, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.venv import installer, installer_options
from pex.venv.install_scope import InstallScope
from pex.venv.installer import Provenance
from pex.venv.installer_configuration import InstallerConfiguration
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Optional


logger = logging.getLogger(__name__)


class InstallLayout(Enum["InstallLayout.Value"]):
    class Value(Enum.Value):
        pass

    VENV = Value("venv")
    FLAT = Value("flat")
    FLAT_ZIPPED = Value("flat-zipped")


class Venv(OutputMixin, JsonMixin, BuildTimeCommand):
    @classmethod
    def _add_inspect_arguments(cls, parser):
        # type: (_ActionsContainer) -> None
        parser.add_argument(
            "venv",
            help="The path of either the venv directory or its Python interpreter.",
        )
        cls.add_output_option(parser, entity="venv information")
        cls.add_json_options(parser, entity="venv information")

    @classmethod
    def _add_create_arguments(cls, parser):
        # type: (_ActionsContainer) -> None
        parser.add_argument(
            "-d",
            "--dir",
            "--dest-dir",
            dest="dest_dir",
            metavar="VENV_DIR",
            required=True,
            help=(
                "The directory to install the venv or flat layout in. If the layout is "
                "{flat_zipped}, then the directory will be installed to and then the zip created "
                "at the same path with a '.zip' extension.".format(
                    flat_zipped=InstallLayout.FLAT_ZIPPED
                )
            ),
        )
        parser.add_argument(
            "--prefix",
            dest="prefix",
            help=(
                "A prefix directory to nest the installation in under the dest dir. This is mainly "
                "useful in the {flat_zipped} layout to inject a fixed prefix to all zip "
                "entries".format(flat_zipped=InstallLayout.FLAT_ZIPPED)
            ),
        )
        parser.add_argument(
            "--layout",
            default=InstallLayout.VENV,
            choices=InstallLayout.values(),
            type=InstallLayout.for_value,
            help=(
                "The layout to create. By default, this is a standard {venv} layout including "
                "activation scripts and a hermetic `sys.path`. The {flat} and {flat_zipped} "
                "layouts can be selected when just the `sys.path` entries are desired. This"
                "effectively exports what would otherwise be the venv `site-packages` directory as "
                "a flat directory that can be joined to the `sys.path` of a compatible"
                "interpreter. These layouts are useful for runtimes that supply an isolated Python "
                "runtime already like AWS Lambda. As a technical detail, these flat layouts "
                "emulate the result of `pip install --target` and include non `site-packages` "
                "installation artifacts at the top level. The common example being a top-level "
                "`bin/` dir containing console scripts.".format(
                    venv=InstallLayout.VENV,
                    flat=InstallLayout.FLAT,
                    flat_zipped=InstallLayout.FLAT_ZIPPED,
                )
            ),
        )
        installer_options.register(parser)
        target_options.register(parser, include_platforms=True)
        resolver_options.register(parser, include_pex_repository=True, include_lock=True)
        requirement_options.register(parser)

    @classmethod
    def add_extra_arguments(
        cls,
        parser,  # type: ArgumentParser
    ):
        # type: (...) -> None
        subcommands = cls.create_subcommands(
            parser,
            description="Interact with virtual environments via the following subcommands.",
        )
        with subcommands.parser(
            name="inspect",
            help="Inspect an existing venv.",
            func=cls._inspect,
            include_verbosity=False,
        ) as inspect_parser:
            cls._add_inspect_arguments(inspect_parser)
        with subcommands.parser(
            name="create",
            help="Create a venv.",
            func=cls._create,
            include_verbosity=True,
        ) as create_parser:
            cls._add_create_arguments(create_parser)

    def _inspect(self):
        # type: () -> Result

        venv = self.options.venv
        if not os.path.exists(venv):
            return Error("The given venv path of {venv} does not exist.".format(venv=venv))

        if os.path.isdir(venv):
            virtualenv = Virtualenv(os.path.normpath(venv))
            if not virtualenv.interpreter.is_venv:
                return Error("{venv} is not a venv.".format(venv=venv))
        else:
            maybe_venv = Virtualenv.enclosing(os.path.normpath(venv))
            if not maybe_venv:
                return Error("{python} is not an venv interpreter.".format(python=venv))
            virtualenv = maybe_venv

        try:
            pex = PexInfo.from_pex(virtualenv.venv_dir)
            is_pex = True
            pex_version = pex.build_properties.get("pex_version")
        except (IOError, OSError, ValueError):
            is_pex = False
            pex_version = None

        venv_info = dict(
            venv_dir=virtualenv.venv_dir,
            provenance=dict(
                created_by=virtualenv.created_by,
                is_pex=is_pex,
                pex_version=pex_version,
            ),
            include_system_site_packages=virtualenv.include_system_site_packages,
            interpreter=dict(
                binary=virtualenv.interpreter.binary,
                base_binary=virtualenv.interpreter.resolve_base_interpreter().binary,
                version=virtualenv.interpreter.identity.version_str,
                sys_path=virtualenv.sys_path,
            ),
            script_dir=virtualenv.bin_dir,
            scripts=sorted(
                os.path.relpath(exe, virtualenv.bin_dir)
                for exe in virtualenv.iter_executables()
                if is_script(exe)
            ),
            site_packages=virtualenv.site_packages_dir,
            distributions=sorted(
                str(dist.as_requirement()) for dist in virtualenv.iter_distributions()
            ),
        )  # type: Dict[str, Any]

        with self.output(self.options) as out:
            self.dump_json(self.options, venv_info, out)
            out.write("\n")

        return Ok()

    def _create(self):
        # type: () -> Result

        targets = target_options.configure(self.options).resolve_targets()
        installer_configuration = installer_options.configure(self.options)

        dest_dir = (
            os.path.join(self.options.dest_dir, self.options.prefix)
            if self.options.prefix
            else self.options.dest_dir
        )
        update = os.path.exists(dest_dir) and not installer_configuration.force
        layout = self.options.layout

        subject = "venv" if layout is InstallLayout.VENV else "flat sys.path directory entry"

        venv = None  # type: Optional[Virtualenv]
        if update and layout is InstallLayout.VENV:
            venv = Virtualenv(venv_dir=dest_dir)
            target = LocalInterpreter.create(venv.interpreter)  # type: Target
            specified_target = try_(
                targets.require_at_most_one_target(
                    purpose="updating venv at {dest_dir}".format(dest_dir=dest_dir)
                )
            )
            if specified_target:
                if specified_target.is_foreign:
                    return Error(
                        "Cannot update a local venv using a foreign platform. Given: "
                        "{platform}.".format(platform=specified_target.platform)
                    )
                original_interpreter = venv.interpreter.resolve_base_interpreter()
                specified_interpreter = (
                    specified_target.get_interpreter().resolve_base_interpreter()
                )
                if specified_interpreter != original_interpreter:
                    return Error(
                        "Cannot update venv at {dest_dir} created with {original_python} using "
                        "{specified_python}".format(
                            dest_dir=dest_dir,
                            original_python=original_interpreter.binary,
                            specified_python=specified_interpreter.binary,
                        )
                    )
            targets = Targets.from_target(target)
        else:
            target = try_(
                targets.require_unique_target(
                    purpose="creating a {subject}".format(subject=subject)
                )
            )
            if layout is InstallLayout.VENV:
                if target.is_foreign:
                    return Error(
                        "Cannot create a local venv for foreign platform {platform}.".format(
                            platform=target.platform
                        )
                    )

                venv = Virtualenv.create(
                    venv_dir=dest_dir,
                    interpreter=target.get_interpreter(),
                    force=installer_configuration.force,
                    copies=installer_configuration.copies,
                    prompt=installer_configuration.prompt,
                )

        requirement_configuration = requirement_options.configure(self.options)
        resolver_configuration = resolver_options.configure(self.options)
        with TRACER.timed("Resolving distributions"):
            installed = configured_resolve.resolve(
                targets=targets,
                requirement_configuration=requirement_configuration,
                resolver_configuration=resolver_configuration,
            )

        pex = None  # type: Optional[PEX]
        lock = None  # type: Optional[str]
        if isinstance(resolver_configuration, PexRepositoryConfiguration):
            pex = PEX(resolver_configuration.pex_repository, interpreter=target.get_interpreter())
        elif isinstance(resolver_configuration, LockRepositoryConfiguration):
            lock = resolver_configuration.lock_file_path

        with TRACER.timed(
            "Installing {count} {wheels} in {subject} at {dest_dir}".format(
                count=len(installed.installed_distributions),
                wheels=pluralize(installed.installed_distributions, "wheel"),
                subject=subject,
                dest_dir=dest_dir,
            )
        ):
            hermetic_scripts = not update and installer_configuration.hermetic_scripts
            distributions = tuple(
                installed_distribution.distribution
                for installed_distribution in installed.installed_distributions
            )
            provenance = (
                Provenance.create(venv=venv)
                if venv
                else Provenance(target_dir=dest_dir, target_python=target.get_interpreter().binary)
            )
            if pex:
                _install_from_pex(
                    pex=pex,
                    installer_configuration=installer_configuration,
                    provenance=provenance,
                    distributions=distributions,
                    dest_dir=dest_dir,
                    hermetic_scripts=hermetic_scripts,
                    venv=venv,
                )
            elif venv:
                installer.populate_venv_distributions(
                    venv=venv,
                    distributions=distributions,
                    provenance=provenance,
                    symlink=False,
                    hermetic_scripts=hermetic_scripts,
                )
            else:
                installer.populate_flat_distributions(
                    dest_dir=dest_dir,
                    distributions=distributions,
                    provenance=provenance,
                    symlink=False,
                )
            source = (
                "PEX at {pex}".format(pex=pex.path())
                if pex
                else "lock at {lock}".format(lock=lock)
                if lock
                else "resolved requirements"
            )
            provenance.check_collisions(
                collisions_ok=installer_configuration.collisions_ok, source=source
            )

        if venv and installer_configuration.pip:
            with TRACER.timed("Installing Pip"):
                try_(
                    installer.ensure_pip_installed(
                        venv,
                        distributions=distributions,
                        scope=installer_configuration.scope,
                        collisions_ok=installer_configuration.collisions_ok,
                        source=source,
                    )
                )

        if installer_configuration.compile:
            with TRACER.timed("Compiling venv sources"):
                try:
                    target.get_interpreter().execute(["-m", "compileall", dest_dir])
                except Executor.NonZeroExit as non_zero_exit:
                    pex_warnings.warn("ignoring compile error {}".format(repr(non_zero_exit)))

        if layout is InstallLayout.FLAT_ZIPPED:
            paths = sorted(
                os.path.join(root, path)
                for root, dirs, files in os.walk(dest_dir)
                for path in itertools.chain(dirs, files)
            )
            unprefixed_dest_dir = self.options.dest_dir
            with open_zip("{dest_dir}.zip".format(dest_dir=unprefixed_dest_dir), "w") as zf:
                for path in paths:
                    zip_entry = zf.zip_entry_from_file(
                        filename=path,
                        arcname=os.path.relpath(path, unprefixed_dest_dir),
                        date_time=DETERMINISTIC_DATETIME.timetuple(),
                    )
                    zf.writestr(zip_entry.info, zip_entry.data)

        return Ok()


def _install_from_pex(
    pex,  # type: PEX
    installer_configuration,  # type: InstallerConfiguration
    provenance,  # type: Provenance
    distributions,  # type: Iterable[Distribution]
    dest_dir,  # type: str
    hermetic_scripts,  # type: bool
    venv=None,  # type: Optional[Virtualenv]
):
    # type: (...) -> None

    if installer_configuration.scope in (InstallScope.ALL, InstallScope.DEPS_ONLY):
        if venv:
            top_level_source_packages = tuple(installer.iter_top_level_source_packages(pex))
            installer.populate_venv_distributions(
                venv=venv,
                distributions=distributions,
                provenance=provenance,
                symlink=False,
                hermetic_scripts=hermetic_scripts,
                top_level_source_packages=top_level_source_packages,
            )
        else:
            installer.populate_flat_distributions(
                dest_dir=dest_dir,
                distributions=distributions,
                provenance=provenance,
                symlink=False,
            )

    if installer_configuration.scope in (InstallScope.ALL, InstallScope.SOURCE_ONLY):
        if venv:
            installer.populate_venv_sources(
                venv=venv,
                pex=pex,
                provenance=provenance,
                bin_path=installer_configuration.bin_path,
                hermetic_scripts=hermetic_scripts,
            )
        else:
            installer.populate_flat_sources(dst=dest_dir, pex=pex, provenance=provenance)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os

try:
    from . import requires_python  # type:ignore[attr-defined] # This file will be relocated.

    python_full_versions = requires_python.PYTHON_FULL_VERSIONS
    python_versions = requires_python.PYTHON_VERSIONS
    python_majors = sorted(set(version[0] for version in python_full_versions))
except ImportError:
    python_full_versions = []
    python_versions = []
    python_majors = []

os_names = []
platform_systems = []
sys_platforms = []
platform_tag_regexps = []

# N.B.: The following environment variables are used by the Pex runtime to control Pip and must be
# kept in-sync with `locker.py`.
target_systems_file = os.environ.pop("_PEX_TARGET_SYSTEMS_FILE", None)

if target_systems_file:
    import json

    with open(target_systems_file) as fp:
        target_systems = json.load(fp)
    os_names = target_systems["os_names"]
    platform_systems = target_systems["platform_systems"]
    sys_platforms = target_systems["sys_platforms"]
    platform_tag_regexps = target_systems["platform_tag_regexps"]


def patch_marker_evaluate():
    from pip._vendor.packaging import markers  # type: ignore[import]

    original_get_env = markers._get_env
    original_eval_op = markers._eval_op

    skip = object()

    def versions_to_string(versions):
        return [".".join(map(str, version)) for version in versions]

    python_versions_strings = versions_to_string(python_versions) or skip
    python_full_versions_strings = versions_to_string(python_full_versions) or skip
    os_names_strings = os_names or skip
    platform_systems_strings = platform_systems or skip
    sys_platforms_strings = sys_platforms or skip

    def _get_env(environment, name):
        if name == "extra":
            return original_get_env(environment, name)
        if name == "python_version":
            return python_versions_strings
        if name == "python_full_version":
            return python_full_versions_strings
        if name == "os_name":
            return os_names_strings
        if name == "platform_system":
            return platform_systems_strings
        if name == "sys_platform":
            return sys_platforms_strings
        return skip

    def _eval_op(lhs, op, rhs):
        if lhs is skip or rhs is skip:
            return True
        return any(
            original_eval_op(left, op, right)
            for left in (lhs if isinstance(lhs, list) else [lhs])
            for right in (rhs if isinstance(rhs, list) else [rhs])
        )

    markers._get_env = _get_env
    markers._eval_op = _eval_op


def patch_wheel_model():
    from pip._internal.models.wheel import Wheel  # type: ignore[import]

    Wheel.support_index_min = lambda *args, **kwargs: 0

    supported_checks = [lambda *args, **kwargs: True]
    if python_versions:
        import re

        def supported_version(self, *_args, **_kwargs):
            if not hasattr(self, "_versions"):
                versions = set()
                is_abi3 = ["abi3"] == list(self.abis)
                for pyversion in self.pyversions:
                    # For the format, see: https://peps.python.org/pep-0425/#python-tag
                    match = re.search(r"^(?P<impl>\D{2,})(?P<major>\d)(?P<minor>\d+)?", pyversion)
                    if not match:
                        continue

                    impl = match.group("impl")
                    if impl not in ("cp", "pp", "py", "cpython", "pypy"):
                        continue

                    major = int(match.group("major"))
                    minor = match.group("minor")
                    if is_abi3 and major == 3:
                        versions.add(major)
                    elif minor:
                        versions.add((major, int(minor)))
                    else:
                        versions.add(major)

                self._versions = versions

            return any(
                (version in python_majors) or (version in python_versions)
                for version in self._versions
            )

        supported_checks.append(supported_version)

    if platform_tag_regexps:
        import re

        def supported_platform_tag(self, *_args, **_kwargs):
            if any(plat == "any" for plat in self.plats):
                return True
            for platform_tag_regexp in platform_tag_regexps:
                if any(re.search(platform_tag_regexp, plat) for plat in self.plats):
                    return True
            return False

        supported_checks.append(supported_platform_tag)

    Wheel.supported = lambda *args, **kwargs: all(
        check(*args, **kwargs) for check in supported_checks
    )

    # N.B.: This patch is a noop for the 20.3.4-patched Pip but is required in newer Pip.
    # The method is used as a speedup hack by newer Pip in some cases instead of
    # Wheel.support_index_min.
    Wheel.find_most_preferred_tag = lambda *args, **kwargs: 0


def patch():
    # 1.) Universal dependency environment marker applicability.
    #
    # Allows all dependencies in metadata to be followed regardless
    # of whether they apply to this system. For example, if this is
    # Python 3.10 but a marker says a dependency is only for
    # 'python_version < "3.6"' we still want to lock that dependency
    # subgraph too.
    patch_marker_evaluate()

    # 2.) Universal wheel tag applicability.
    #
    # Allows all wheel URLs to be checked even when the wheel does not
    # match system tags.
    patch_wheel_model()



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json

from pex import compatibility
from pex.dist_metadata import Requirement, RequirementParseError
from pex.enum import Enum
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.locked_resolve import (
    Artifact,
    LockedRequirement,
    LockedResolve,
    LockStyle,
    TargetSystem,
)
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.path_mappings import PathMappings
from pex.resolve.resolved_requirement import Fingerprint, Pin
from pex.resolve.resolver_configuration import PipConfiguration, ResolverVersion
from pex.sorted_tuple import SortedTuple
from pex.third_party.packaging import tags
from pex.third_party.packaging.specifiers import InvalidSpecifier, SpecifierSet
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import (
        Any,
        Container,
        Dict,
        List,
        Mapping,
        Optional,
        Text,
        Tuple,
        Type,
        TypeVar,
        Union,
    )

    import attr  # vendor:skip

    _V = TypeVar("_V", bound=Enum.Value)
else:
    from pex.third_party import attr


class ParseError(Exception):
    """Indicates an error parsing a Pex lock file."""


@attr.s(frozen=True)
class PathMappingError(ParseError):
    """Indicates missing path mappings when parsing a Pex lock file."""

    required_path_mappings = attr.ib()  # type: Mapping[str, Optional[str]]
    unspecified_paths = attr.ib()  # type: Container[str]


def _load_json(
    lockfile_contents,  # type: Text
    source,  # type: str
):
    # type: (...) -> Mapping
    try:
        return cast("Mapping", json.loads(lockfile_contents))
    except ValueError as e:
        raise ParseError(
            "The lock file at {source} does not contain valid JSON: "
            "{err}".format(source=source, err=e)
        )


_DEFAULT_PIP_CONFIGURATION = PipConfiguration()


def loads(
    lockfile_contents,  # type: Text
    source="<string>",  # type: str
    path_mappings=PathMappings(),  # type: PathMappings
):
    # type: (...) -> Lockfile

    def get(
        key,  # type: str
        expected_type=compatibility.string,  # type: Union[Type, Tuple[Type, ...]]
        data=_load_json(lockfile_contents, source=source),  # type: Mapping
        path=".",  # type: str
        optional=False,
    ):
        # type: (...) -> Any

        if not isinstance(data, dict):
            raise ParseError(
                "Cannot retrieve '{path}[\"{key}\"]' in {source} because '{path}' is not a "
                "JSON object but a {type} with value {value}.".format(
                    path=path,
                    key=key,
                    source=source,
                    type=type(data).__name__,
                    value=data,
                )
            )
        if key not in data:
            if optional:
                return None
            raise ParseError(
                "The object at '{path}' in {source} did not have the expected key "
                "{key!r}.".format(path=path, source=source, key=key)
            )
        value = data[key]
        if not optional and not isinstance(value, expected_type):
            raise ParseError(
                "Expected '{path}[\"{key}\"]' in {source} to be of type {expected_type} "
                "but given {type} with value {value}.".format(
                    path=path,
                    key=key,
                    source=source,
                    expected_type=(
                        " or ".join(t.__name__ for t in expected_type)
                        if isinstance(expected_type, tuple)
                        else expected_type.__name__
                    ),
                    type=type(value).__name__,
                    value=value,
                )
            )
        return value

    def parse_enum_value(
        enum_type,  # type: Type[Enum[_V]]
        value,  # type: str
        path=".",  # type: str
    ):
        # type: (...) -> _V
        try:
            return enum_type.for_value(value)
        except ValueError as e:
            raise ParseError("The '{path}' is invalid: {err}".format(path=path, err=e))

    def get_enum_value(
        enum_type,  # type: Type[Enum[_V]]
        key,  # type: str
        path=".",  # type: str
        default=None,  # type: Optional[_V]
    ):
        # type: (...) -> _V
        value = get(key, path=path, optional=default is not None)
        if not value and default:
            return default
        return parse_enum_value(
            enum_type=enum_type, value=value, path='{path}["{key}"]'.format(path=path, key=key)
        )

    def parse_requirement(
        raw_requirement,  # type: str
        path,  # type: str
    ):
        # type: (...) -> Requirement
        try:
            return Requirement.parse(path_mappings.maybe_reify(raw_requirement))
        except RequirementParseError as e:
            raise ParseError(
                "The requirement string at '{path}' is invalid: {err}".format(path=path, err=e)
            )

    def parse_version_specifier(
        raw_version_specifier,  # type: str
        path,  # type: str
    ):
        # type: (...) -> SpecifierSet
        try:
            return SpecifierSet(raw_version_specifier)
        except InvalidSpecifier as e:
            raise ParseError(
                "The version specifier at '{path}' is invalid: {err}".format(path=path, err=e)
            )

    required_path_mappings = get("path_mappings", dict, optional=True) or {}
    given_mappings = set(mapping.name for mapping in path_mappings.mappings)
    unspecified_paths = set(required_path_mappings) - given_mappings
    if unspecified_paths:
        raise PathMappingError(
            required_path_mappings=required_path_mappings, unspecified_paths=unspecified_paths
        )

    target_systems = [
        parse_enum_value(
            enum_type=TargetSystem,
            value=target_system,
            path=".target_systems[{index}]".format(index=index),
        )
        for index, target_system in enumerate(get("target_systems", list, optional=True) or ())
    ]

    requirements = [
        parse_requirement(req, path=".requirements[{index}]".format(index=index))
        for index, req in enumerate(get("requirements", list))
    ]

    constraints = [
        parse_requirement(constraint, path=".constraints[{index}]".format(index=index))
        for index, constraint in enumerate(get("constraints", list))
    ]

    def assemble_tag(
        components,  # type: List[str]
        path,  # type: str
    ):
        # type: (...) -> tags.Tag
        if len(components) != 3 or not all(isinstance(c, compatibility.string) for c in components):
            raise ParseError(
                "The tag at '{path}' must have 3 string components. Given {count} with types "
                "[{types}]: {components}".format(
                    path=path,
                    count=len(components),
                    types=", ".join(type(c).__name__ for c in components),
                    components=components,
                )
            )
        return tags.Tag(interpreter=components[0], abi=components[1], platform=components[2])

    locked_resolves = []
    for lock_index, locked_resolve in enumerate(get("locked_resolves", list)):
        lock_path = ".locked_resolves[{lock_index}]".format(lock_index=lock_index)
        platform_tag_components = get(
            "platform_tag", list, data=locked_resolve, path=lock_path, optional=True
        )
        platform_tag = (
            assemble_tag(
                components=platform_tag_components,
                path='{lock_path}["platform_tag"]'.format(lock_path=lock_path),
            )
            if platform_tag_components
            else None
        )
        locked_reqs = []
        for req_index, req in enumerate(
            get("locked_requirements", list, data=locked_resolve, path=lock_path)
        ):
            req_path = "{lock_path}[{req_index}]".format(lock_path=lock_path, req_index=req_index)

            artifacts = []
            for i, artifact in enumerate(get("artifacts", list, data=req, path=req_path)):
                ap = '{path}["artifacts"][{index}]'.format(path=req_path, index=i)
                artifacts.append(
                    Artifact.from_url(
                        url=path_mappings.maybe_reify(get("url", data=artifact, path=ap)),
                        fingerprint=Fingerprint(
                            algorithm=get("algorithm", data=artifact, path=ap),
                            hash=get("hash", data=artifact, path=ap),
                        ),
                    )
                )

            if not artifacts:
                raise ParseError(
                    "Expected '{path}' in {source} to have at least one artifact.".format(
                        path=req_path, source=source
                    )
                )

            requires_python = None
            version_specifier = get("requires_python", data=req, path=req_path, optional=True)
            if version_specifier:
                requires_python = parse_version_specifier(
                    version_specifier, path='{path}["requires_python"]'.format(path=req_path)
                )

            locked_reqs.append(
                LockedRequirement.create(
                    pin=Pin(
                        project_name=ProjectName(get("project_name", data=req, path=req_path)),
                        version=Version(get("version", data=req, path=req_path)),
                    ),
                    requires_python=requires_python,
                    requires_dists=[
                        parse_requirement(
                            requires_dist,
                            path='{path}["requires_dists"][{index}]'.format(path=req_path, index=i),
                        )
                        for i, requires_dist in enumerate(
                            get("requires_dists", list, data=req, path=req_path)
                        )
                    ],
                    artifact=artifacts[0],
                    additional_artifacts=artifacts[1:],
                )
            )

        locked_resolves.append(
            LockedResolve(locked_requirements=SortedTuple(locked_reqs), platform_tag=platform_tag)
        )

    return Lockfile.create(
        pex_version=get("pex_version"),
        style=get_enum_value(LockStyle, "style"),
        requires_python=get("requires_python", list),
        target_systems=target_systems,
        pip_version=get_enum_value(
            PipVersion, "pip_version", default=_DEFAULT_PIP_CONFIGURATION.version
        ),
        resolver_version=get_enum_value(ResolverVersion, "resolver_version"),
        requirements=requirements,
        constraints=constraints,
        allow_prereleases=get("allow_prereleases", bool),
        allow_wheels=get("allow_wheels", bool),
        allow_builds=get("allow_builds", bool),
        prefer_older_binary=get("prefer_older_binary", bool),
        use_pep517=get("use_pep517", bool, optional=True),
        build_isolation=get("build_isolation", bool),
        transitive=get("transitive", bool),
        locked_resolves=locked_resolves,
        source=source,
    )


def load(
    lockfile_path,  # type: str
    path_mappings=PathMappings(),  # type: PathMappings
):
    # type: (...) -> Lockfile
    try:
        with open(lockfile_path) as fp:
            return loads(fp.read(), source=lockfile_path, path_mappings=path_mappings)
    except IOError as e:
        raise ParseError(
            "Failed to read lock file at {path}: {err}".format(path=lockfile_path, err=e)
        )


def as_json_data(
    lockfile,  # type: Lockfile
    path_mappings=PathMappings(),  # type: PathMappings
):
    # type: (...) -> Dict[str, Any]
    return {
        "pex_version": lockfile.pex_version,
        "style": str(lockfile.style),
        "requires_python": list(lockfile.requires_python),
        "target_systems": [str(target_system) for target_system in lockfile.target_systems],
        "pip_version": str(lockfile.pip_version),
        "resolver_version": str(lockfile.resolver_version),
        "requirements": [
            path_mappings.maybe_canonicalize(str(req)) for req in lockfile.requirements
        ],
        "constraints": [str(constraint) for constraint in lockfile.constraints],
        "allow_prereleases": lockfile.allow_prereleases,
        "allow_wheels": lockfile.allow_wheels,
        "allow_builds": lockfile.allow_builds,
        "prefer_older_binary": lockfile.prefer_older_binary,
        "use_pep517": lockfile.use_pep517,
        "build_isolation": lockfile.build_isolation,
        "transitive": lockfile.transitive,
        "locked_resolves": [
            {
                "platform_tag": [
                    locked_resolve.platform_tag.interpreter,
                    locked_resolve.platform_tag.abi,
                    locked_resolve.platform_tag.platform,
                ]
                if locked_resolve.platform_tag
                else None,
                "locked_requirements": [
                    {
                        "project_name": str(req.pin.project_name),
                        # N.B.: We store the raw version so that `===` can work as intended against
                        # the un-normalized form of versions that are non-legacy and thus
                        # normalizable.
                        "version": req.pin.version.raw,
                        "requires_dists": [
                            path_mappings.maybe_canonicalize(str(dependency))
                            for dependency in req.requires_dists
                        ],
                        "requires_python": str(req.requires_python)
                        if req.requires_python
                        else None,
                        "artifacts": [
                            {
                                "url": path_mappings.maybe_canonicalize(artifact.url),
                                "algorithm": artifact.fingerprint.algorithm,
                                "hash": artifact.fingerprint.hash,
                            }
                            for artifact in req.iter_artifacts()
                        ],
                    }
                    for req in locked_resolve.locked_requirements
                ],
            }
            for locked_resolve in lockfile.locked_resolves
        ],
        "path_mappings": {
            path_mapping.name: path_mapping.description for path_mapping in path_mappings.mappings
        },
    }



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import re
from netrc import NetrcParseError, netrc

from pex import pex_warnings
from pex.compatibility import urlparse
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Optional, Text, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Machine(object):
    @classmethod
    def from_url_info(cls, url_info):
        # type: (urlparse.ParseResult) -> Machine

        if not url_info.scheme:
            raise ValueError(
                "Expected a scheme for the BaseURL. Given: {url_info}".format(url_info=url_info)
            )

        if "file" == url_info.scheme:
            return cls(reduced_url="file://{path}".format(path=url_info.path))

        if not url_info.hostname:
            raise ValueError(
                "Expected a hostname for a BaseURL with {scheme} scheme. Given: {url_info}".format(
                    scheme=url_info.scheme, url_info=url_info
                )
            )

        return cls(
            reduced_url="{scheme}://{host}{port}".format(
                scheme=url_info.scheme,
                host=url_info.hostname,
                port=":{port}".format(port=url_info.port) if url_info.port is not None else "",
            ),
            host=url_info.hostname,
            port=url_info.port,
        )

    @classmethod
    def from_url(cls, url):
        # type: (Text) -> Machine
        return cls.from_url_info(urlparse.urlparse(url))

    reduced_url = attr.ib()  # type: str
    host = attr.ib(default=None)  # type: Optional[str]
    port = attr.ib(default=None)  # type: Optional[int]


@attr.s(frozen=True)
class PasswordEntry(object):
    @classmethod
    def maybe_extract_from_url(cls, url):
        # type: (str) -> Optional[PasswordEntry]
        url_info = urlparse.urlparse(url)
        if not url_info.username or not url_info.password:
            return None

        return cls(
            machine=Machine.from_url_info(url_info),
            username=url_info.username,
            password=url_info.password,
        )

    username = attr.ib()  # type: str
    password = attr.ib()  # type: str
    machine = attr.ib(default=None)  # type: Optional[Machine]

    def uri_or_default(self, url):
        # type: (Text) -> str
        return (self.machine or Machine.from_url(url)).reduced_url

    def maybe_inject_in_url(self, url):
        # type: (str) -> Optional[str]
        url_info = urlparse.urlparse(url)

        if url_info.username or url_info.password:
            # Don't stomp credentials already embedded in an URL.
            return None

        if self.machine:
            if Machine.from_url_info(url_info) != self.machine:
                return None

        scheme, netloc, path, params, query, fragment = url_info
        netloc = "{username}:{password}@{netloc}".format(
            username=self.username, password=self.password, netloc=netloc
        )
        return urlparse.urlunparse((scheme, netloc, path, params, query, fragment))


@attr.s(frozen=True)
class PasswordDatabase(object):
    @classmethod
    def from_netrc(cls, netrc_file="~/.netrc"):
        # type: (Optional[str]) -> PasswordDatabase
        if not netrc_file:
            return cls()

        netrc_path = os.path.expanduser(netrc_file)
        if not os.path.isfile(netrc_path):
            return cls()

        try:
            netrc_database = netrc(netrc_path)
        except NetrcParseError as e:
            pex_warnings.warn(
                "Failed to load netrc credentials: {err}\n"
                "Continuing without netrc credentials.".format(err=e)
            )
            return cls()

        def iter_entries():
            # type: () -> Iterator[PasswordEntry]
            for machine, (login, _, password) in netrc_database.hosts.items():
                if password is None:
                    continue

                if machine == "default":
                    # The `default` entry is special and means just that; so we omit any qualifying
                    # machine.
                    yield PasswordEntry(username=login, password=password)

                # Traditionally, ~/.netrc machine entries just contain a bare hostname but we allow
                # for them containing a scheme since the format is so poorly documented and really
                # used as the beholder sees fit (it's a shared ~community format at this point and
                # used differently by a wide range of tools; I think FTP originally - thus it has
                # `macros` entries which ~no-one knows about / uses).
                #
                # For scheme syntax, see: https://datatracker.ietf.org/doc/html/rfc3986#section-3.1
                if re.search(r"^(?P<scheme>[a-zA-Z][a-zA-Z\d+.-]*)://", machine):
                    yield PasswordEntry(
                        machine=Machine.from_url(machine),
                        username=login,
                        password=password,
                    )
                    continue

                for scheme in "http", "https":
                    yield PasswordEntry(
                        machine=Machine.from_url(
                            "{scheme}://{machine}".format(scheme=scheme, machine=machine)
                        ),
                        username=login,
                        password=password,
                    )

        return cls(entries=tuple(iter_entries()))

    entries = attr.ib(default=())  # type: Tuple[PasswordEntry, ...]

    def append(self, entries):
        # type: (Iterable[PasswordEntry]) -> PasswordDatabase
        return PasswordDatabase(entries=self.entries + tuple(entries))



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os
import subprocess
import sys
from textwrap import dedent
from typing import Text, Tuple

import pytest

from pex.common import safe_open
from pex.testing import (
    ALL_PY_VERSIONS,
    all_pythons,
    ensure_python_interpreter,
    make_env,
    run_pex_command,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, List


@pytest.mark.parametrize(
    ["execution_mode_args"],
    [
        pytest.param([], id="zipapp"),
        pytest.param(["--venv"], id="venv"),
    ],
)
def test_execute(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=["cowsay==4.0", "-c", "cowsay", "-o", cowsay, "--sh-boot"] + execution_mode_args
    ).assert_success()
    assert "4.0" == subprocess.check_output(args=[cowsay, "--version"]).decode("utf-8").strip()


@pytest.mark.parametrize(
    ["execution_mode_args"],
    [
        pytest.param([], id="zipapp"),
        pytest.param(["--venv"], id="venv"),
    ],
)
def test_issue_1881(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    os.mkdir(pex_root)
    # Test that the runtime_pex_root is respected even when it is unwritable at build time.
    os.chmod(pex_root, 0o555)
    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=[
            "cowsay==4.0",
            "-c",
            "cowsay",
            "-o",
            cowsay,
            "--python-shebang",
            sys.executable,
            "--sh-boot",
            "--runtime-pex-root",
            pex_root,
        ]
        + execution_mode_args
    ).assert_success()
    # simulate pex_root writable at runtime.
    os.chmod(pex_root, 0o777)

    def _execute_cowsay_pex():
        return (
            subprocess.check_output(
                args=[cowsay, "--version"], env=make_env(PEX_VERBOSE=1), stderr=subprocess.STDOUT
            )
            .decode("utf-8")
            .strip()
        )

    # When this string is logged from the sh_boot script it indicates that the slow
    # path of running the zipapp via python interpreter taken.
    slow_path_output = "Running zipapp pex to lay itself out under PEX_ROOT."
    assert slow_path_output in _execute_cowsay_pex()
    assert slow_path_output not in _execute_cowsay_pex()


def interpreters():
    # type: () -> Iterable[Tuple[Text, List[Text]]]

    def iter_interpreters():
        # type: () -> Iterator[Tuple[Text, List[Text]]]

        def entry(path):
            # type: (Text) -> Tuple[Text, List[Text]]
            return os.path.basename(path), [path]

        yield entry(sys.executable)

        for interpreter in all_pythons():
            yield entry(interpreter)

        locations = (
            subprocess.check_output(
                args=["/usr/bin/env", "bash", "-c", "command -v ash bash busybox dash ksh sh zsh"]
            )
            .decode("utf-8")
            .splitlines()
        )
        for location in locations:
            basename = os.path.basename(location)
            if "busybox" == basename:
                yield "ash (via busybox)", [location, "ash"]
            else:
                yield entry(location)

    return sorted({name: args for name, args in iter_interpreters()}.items())


@pytest.mark.parametrize(
    ["interpreter_cmd"],
    [pytest.param(args, id=name) for name, args in interpreters()],
)
def test_execute_via_interpreter(
    tmpdir,  # type: Any
    interpreter_cmd,  # type: List[str]
):
    # type: (...) -> None

    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=["cowsay==4.0", "-c", "cowsay", "-o", cowsay, "--sh-boot"]
    ).assert_success()

    assert (
        "4.0"
        == subprocess.check_output(args=interpreter_cmd + [cowsay, "--version"])
        .decode("utf-8")
        .strip()
    )


def test_python_shebang_respected(tmpdir):
    # type: (Any) -> None

    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=[
            "cowsay==4.0",
            "-c",
            "cowsay",
            "-o",
            cowsay,
            "--sh-boot",
            "--python-shebang",
            # This is a strange shebang ~no-one would use since it short-circuits the PEX execution
            # to always just print the Python interpreter version, but it serves the purposes of:
            # 1. Proving our python shebang is honored by the bash boot.
            # 2. The bash boot treatment can handle shebangs with arguments in them.
            "{python} -V".format(python=sys.executable),
        ]
    ).assert_success()

    # N.B.: Python 2.7 does not send version to stdout; so we redirect stdout to stderr to be able
    # to uniformly retrieve the Python version.
    output = subprocess.check_output(args=[cowsay], stderr=subprocess.STDOUT).decode("utf-8")
    version = "Python {version}".format(version=".".join(map(str, sys.version_info[:3])))
    assert output.startswith(version), output


EXECUTION_MODE_ARGS_PERMUTATIONS = [
    pytest.param([], id="ZIPAPP"),
    pytest.param(["--venv"], id="VENV"),
    pytest.param(["--sh-boot"], id="ZIPAPP (--sh-boot)"),
    pytest.param(["--venv", "--sh-boot"], id="VENV (--sh-boot)"),
]


@pytest.mark.parametrize("execution_mode_args", EXECUTION_MODE_ARGS_PERMUTATIONS)
def test_issue_1782(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex.sh"))
    run_pex_command(
        args=[pex_project_dir, "-c", "pex", "-o", pex] + execution_mode_args
    ).assert_success()

    help_line1 = subprocess.check_output(args=[pex, "-h"]).decode("utf-8").splitlines()[0]
    assert help_line1.startswith("usage: {pex} ".format(pex=os.path.basename(pex))), help_line1
    assert (
        pex
        == subprocess.check_output(
            args=[pex, "-c", "import os; print(os.environ['PEX'])"], env=make_env(PEX_INTERPRETER=1)
        )
        .decode("utf-8")
        .strip()
    )


@pytest.mark.parametrize("execution_mode_args", EXECUTION_MODE_ARGS_PERMUTATIONS)
def test_argv0(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex.sh"))
    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "app.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import json
                import os
                import sys


                def main():
                    print(json.dumps({"PEX": os.environ["PEX"], "argv0": sys.argv[0]}))


                if __name__ == "__main__":
                    main()
                """
            )
        )

    run_pex_command(
        args=["-D", src, "-e", "app:main", "-o", pex] + execution_mode_args
    ).assert_success()
    assert {"PEX": pex, "argv0": pex} == json.loads(subprocess.check_output(args=[pex]))

    run_pex_command(args=["-D", src, "-m", "app", "-o", pex] + execution_mode_args).assert_success()
    data = json.loads(subprocess.check_output(args=[pex]))
    assert pex == data.pop("PEX")
    assert "app.py" == os.path.basename(data.pop("argv0")), (
        "When executing modules we expect runpy.run_module to `alter_sys` in order to support "
        "pickling and other use cases as outlined in https://github.com/pantsbuild/pex/issues/1018."
    )
    assert {} == data



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
from collections import OrderedDict

from pex.common import pluralize
from pex.network_configuration import NetworkConfiguration
from pex.orderedset import OrderedSet
from pex.requirements import LocalProjectRequirement, parse_requirement_strings
from pex.resolve.locked_resolve import DownloadableArtifact, Resolved
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.result import Error
from pex.targets import Target, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Iterable, Optional, Tuple, Union

    import attr  # vendor:skip

    from pex.requirements import ParsedRequirement
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Subset(object):
    target = attr.ib()  # type: Target
    resolved = attr.ib()  # type: Resolved


@attr.s(frozen=True)
class SubsetResult(object):
    requirements = attr.ib()  # type: Tuple[ParsedRequirement, ...]
    subsets = attr.ib()  # type: Tuple[Subset, ...]


def subset(
    targets,  # type: Targets
    lock,  # type: Lockfile
    requirement_configuration=RequirementConfiguration(),  # type: RequirementConfiguration
    network_configuration=None,  # type: Optional[NetworkConfiguration]
    build=True,  # type: bool
    use_wheel=True,  # type: bool
    prefer_older_binary=False,  # type: bool
    transitive=True,  # type: bool
    include_all_matches=False,  # type: bool
):
    # type: (...) -> Union[SubsetResult, Error]

    with TRACER.timed("Parsing requirements"):
        parsed_requirements = tuple(
            requirement_configuration.parse_requirements(network_configuration)
        ) or tuple(parse_requirement_strings(str(req) for req in lock.requirements))
        constraints = tuple(
            parsed_constraint.requirement
            for parsed_constraint in requirement_configuration.parse_constraints(
                network_configuration
            )
        )
        requirements_to_resolve = OrderedSet(
            lock.local_project_requirement_mapping[os.path.abspath(parsed_requirement.path)]
            if isinstance(parsed_requirement, LocalProjectRequirement)
            else parsed_requirement.requirement
            for parsed_requirement in parsed_requirements
        )

    resolved_by_target = OrderedDict()  # type: OrderedDict[Target, Resolved]
    errors_by_target = {}  # type: Dict[Target, Iterable[Error]]

    with TRACER.timed(
        "Resolving urls to fetch for {count} requirements from lock {lockfile}".format(
            count=len(parsed_requirements), lockfile=lock.source
        )
    ):
        for target in targets.unique_targets():
            resolveds = []
            errors = []
            for locked_resolve in lock.locked_resolves:
                resolve_result = locked_resolve.resolve(
                    target,
                    requirements_to_resolve,
                    constraints=constraints,
                    source=lock.source,
                    build=build,
                    use_wheel=use_wheel,
                    prefer_older_binary=prefer_older_binary,
                    transitive=transitive,
                    include_all_matches=include_all_matches,
                    # TODO(John Sirois): Plumb `--ignore-errors` to support desired but technically
                    #  invalid `pip-legacy-resolver` locks:
                    #  https://github.com/pantsbuild/pex/issues/1652
                )
                if isinstance(resolve_result, Resolved):
                    resolveds.append(resolve_result)
                else:
                    errors.append(resolve_result)

            if resolveds:
                resolved_by_target[target] = Resolved.most_specific(resolveds)
            elif errors:
                errors_by_target[target] = tuple(errors)

    if errors_by_target:
        return Error(
            "Failed to resolve compatible artifacts from {lock} for {count} {targets}:\n"
            "{errors}".format(
                lock="lock {source}".format(source=lock.source) if lock.source else "lock",
                count=len(errors_by_target),
                targets=pluralize(errors_by_target, "target"),
                errors="\n".join(
                    "{index}. {target}:\n    {errors}".format(
                        index=index, target=target, errors="\n    ".join(map(str, errors))
                    )
                    for index, (target, errors) in enumerate(errors_by_target.items(), start=1)
                ),
            )
        )

    return SubsetResult(
        requirements=parsed_requirements,
        subsets=tuple(
            Subset(target=target, resolved=resolved)
            for target, resolved in resolved_by_target.items()
        ),
    )



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.enum import Enum
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union


class InheritPath(Enum["InheritPath.Value"]):
    class Value(Enum.Value):
        pass

    FALSE = Value("false")
    PREFER = Value("prefer")
    FALLBACK = Value("fallback")

    @classmethod
    def for_value(cls, value):
        # type: (Union[str, bool]) -> InheritPath.Value
        if not isinstance(value, bool):
            return super(InheritPath, cls).for_value(value)
        elif value is False:
            return InheritPath.FALSE
        elif value is True:
            return InheritPath.PREFER
        else:
            raise ValueError(
                "An InheritPath.Value must be a str or a bool; given {} of type {}".format(
                    value, type(value)
                )
            )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.fetcher import URLFetcher
from pex.network_configuration import NetworkConfiguration
from pex.requirements import Constraint, parse_requirement_file, parse_requirement_strings
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, List, Optional

    import attr  # vendor:skip

    from pex.requirements import ParsedRequirement
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class RequirementConfiguration(object):
    requirements = attr.ib(default=None)  # type: Optional[Iterable[str]]
    requirement_files = attr.ib(default=None)  # type: Optional[Iterable[str]]
    constraint_files = attr.ib(default=None)  # type: Optional[Iterable[str]]

    def parse_requirements(self, network_configuration=None):
        # type: (Optional[NetworkConfiguration]) -> Iterable[ParsedRequirement]
        parsed_requirements = []  # type: List[ParsedRequirement]
        if self.requirements:
            parsed_requirements.extend(parse_requirement_strings(self.requirements))
        if self.requirement_files:
            fetcher = URLFetcher(network_configuration=network_configuration)
            for requirement_file in self.requirement_files:
                parsed_requirements.extend(
                    requirement_or_constraint
                    for requirement_or_constraint in parse_requirement_file(
                        requirement_file, is_constraints=False, fetcher=fetcher
                    )
                    if not isinstance(requirement_or_constraint, Constraint)
                )
        return parsed_requirements

    def parse_constraints(self, network_configuration=None):
        # type: (Optional[NetworkConfiguration]) -> Iterable[Constraint]
        parsed_constraints = []  # type: List[Constraint]
        if self.constraint_files:
            fetcher = URLFetcher(network_configuration=network_configuration)
            for constraint_file in self.constraint_files:
                parsed_constraints.extend(
                    requirement_or_constraint
                    for requirement_or_constraint in parse_requirement_file(
                        constraint_file, is_constraints=True, fetcher=fetcher
                    )
                    if isinstance(requirement_or_constraint, Constraint)
                )
        return parsed_constraints



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

from pex.pex_info import PexInfo
from pex.testing import run_pex_command, run_simple_pex
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Any, Dict, Tuple


def test_venv_mode_dir_hash_includes_all_pex_info_metadata(tmpdir):
    # type: (Any) -> None

    def get_fabric_versions(pex):
        # type: (str) -> Dict[str, str]
        output, returncode = run_simple_pex(pex, args=["--version"])
        assert 0 == returncode
        return dict(
            cast("Tuple[str, str]", line.split(" ", 1))
            for line in output.decode("utf-8").splitlines()
        )

    # The only difference in these two PEX files is their entrypoint. Ensure venv execution takes
    # that into account and disambiguates the otherwise identical PEX files.

    invoke_pex = os.path.join(str(tmpdir), "invoke.pex")
    results = run_pex_command(
        args=["fabric==2.6.0", "invoke==1.5.0", "--venv", "-e", "invoke", "-o", invoke_pex],
        quiet=True,
    )
    results.assert_success()
    invoke_versions = get_fabric_versions(invoke_pex)
    assert len(invoke_versions) == 1
    invoke_version = invoke_versions["Invoke"]
    assert invoke_version == "1.5.0"

    fabric_pex = os.path.join(str(tmpdir), "fabric.pex")
    results = run_pex_command(
        args=[
            "fabric==2.6.0",
            "--venv",
            "-e",
            "fabric",
            "-o",
            fabric_pex,
            "--pex-repository",
            invoke_pex,
        ],
        quiet=True,
    )
    results.assert_success()
    fabric_versions = get_fabric_versions(fabric_pex)
    assert len(fabric_versions) >= 2
    assert invoke_version == fabric_versions["Invoke"]
    assert "2.6.0" == fabric_versions["Fabric"]

    invoke_pex_info = PexInfo.from_pex(invoke_pex)
    fabric_pex_info = PexInfo.from_pex(fabric_pex)
    assert invoke_pex_info.code_hash == fabric_pex_info.code_hash
    assert invoke_pex_info.distributions == fabric_pex_info.distributions
    assert invoke_pex_info.pex_hash != fabric_pex_info.pex_hash



#!/usr/bin/env python3

import hashlib
import io
import os
import subprocess
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from enum import Enum, unique
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path, PurePath
from typing import Optional, Tuple, cast

PROJECT_METADATA = Path("pyproject.toml")
DIST_DIR = Path("dist")


def python_requires() -> str:
    # N.B.: We run this script directly from integration tests where toml is not needed; so we lazy
    # import to allow for this direct running of the script without dependency installation by tox
    # in ITs.
    import toml

    project_metadata = toml.loads(PROJECT_METADATA.read_text())
    return cast(str, project_metadata["tool"]["flit"]["metadata"]["requires-python"].strip())


def build_pex_pex(output_file: PurePath, local: bool = False, verbosity: int = 0) -> None:
    # NB: We do not include the subprocess extra (which would be spelled: `.[subprocess]`) since we
    # would then produce a pex that would not be consumable by all python interpreters otherwise
    # meeting `python_requires`; ie: we'd need to then come up with a deploy environment / deploy
    # tooling, that built subprocess32 for linux cp27m, cp27mu, pypy, ... etc. Even with all the
    # work expended to do this, we'd still miss some platform someone wanted to run the Pex PEX on.
    # As such, we just ship unadorned Pex which is pure-python and universal. Any user wanting the
    # extra is encouraged to build a Pex PEX for their particular platform themselves.
    pex_requirement = "."

    args = [
        sys.executable,
        "-m",
        "pex",
        *["-v" for _ in range(verbosity)],
        "--disable-cache",
        "--no-build",
        "--no-compile",
        "--no-use-system-time",
        "--python-shebang",
        "/usr/bin/env python",
        "--no-strip-pex-env",
        "--include-tools",
        "-o",
        str(output_file),
        "-c",
        "pex",
        pex_requirement,
    ]
    if not local:
        args.extend(["--interpreter-constraint", python_requires()])
    subprocess.run(args, check=True)


def describe_git_rev() -> str:
    git_describe = subprocess.run(
        ["git", "describe"], check=True, capture_output=True, encoding="utf-8"
    )
    return git_describe.stdout.strip()


def describe_file(path: Path) -> Tuple[str, int]:
    hasher = hashlib.sha256()
    size = 0
    with path.open("rb") as fp:
        for chunk in iter(lambda: fp.read(io.DEFAULT_BUFFER_SIZE), b""):
            hasher.update(chunk)
            size += len(chunk)

    return hasher.hexdigest(), size


@unique
class Format(Enum):
    SDIST = "sdist"
    WHEEL = "wheel"

    def __str__(self) -> str:
        return cast(str, self.value)


def build_pex_dists(dist_fmt: Format, *additional_dist_fmts: Format, verbose: bool = False) -> None:
    output = None if verbose else subprocess.DEVNULL
    subprocess.run(
        ["flit", "build", *[f"--format={fmt}" for fmt in [dist_fmt, *additional_dist_fmts]]],
        stdout=output,
        stderr=output,
        check=True,
    )


def main(
    *additional_dist_formats: Format,
    verbosity: int = 0,
    pex_output_file: Optional[Path] = DIST_DIR / "pex",
    local: bool = False,
    serve: bool = False
) -> None:
    if pex_output_file:
        print(f"Building Pex PEX to `{pex_output_file}` ...")
        build_pex_pex(pex_output_file, local, verbosity)

        git_rev = describe_git_rev()
        sha256, size = describe_file(pex_output_file)
        print(f"Built Pex PEX @ {git_rev}:")
        print(f"sha256: {sha256}")
        print(f"  size: {size}")

    if additional_dist_formats:
        print(
            f"Building additional distribution formats to `{DIST_DIR}`: "
            f'{", ".join(f"{i + 1}.) {fmt}" for i, fmt in enumerate(additional_dist_formats))} ...'
        )
        build_pex_dists(*additional_dist_formats, verbose=verbosity > 0)
        print("Built:")
        for root, _, files in os.walk(DIST_DIR):
            root_path = Path(root)
            for f in files:
                dist_path = root_path / f
                if dist_path != pex_output_file:
                    print(f"  {dist_path}")

    if serve:
        server = HTTPServer(("", 0), SimpleHTTPRequestHandler)
        host, port = server.server_address

        print(f"Serving Pex distributions from `{DIST_DIR}` at http://{host}:{port} ...")

        os.chdir(DIST_DIR)
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            print(f"Server shut down in response to keyboard interrupt.")


if __name__ == "__main__":
    if not PROJECT_METADATA.is_file():
        print("This script must be run from the root of the Pex repo.", file=sys.stderr)
        sys.exit(1)

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-v", dest="verbosity", action="count", default=0, help="Increase output verbosity level."
    )
    parser.add_argument(
        "--additional-format",
        dest="additional_formats",
        choices=list(Format),
        type=Format,
        action="append",
        help="Package Pex in additional formats.",
    )
    parser.add_argument(
        "--no-pex",
        default=False,
        action="store_true",
        help="Do not build the Pex PEX.",
    )
    parser.add_argument(
        "--pex-output-file",
        default=DIST_DIR / "pex",
        type=Path,
        help="Build the Pex PEX at this path.",
    )
    parser.add_argument(
        "--local",
        default=False,
        action="store_true",
        help="Build Pex PEX with just a single local interpreter.",
    )
    parser.add_argument(
        "--serve",
        default=False,
        action="store_true",
        help="After packaging Pex serve up the packages over HTTP.",
    )
    args = parser.parse_args()

    main(
        *(args.additional_formats or ()),
        verbosity=args.verbosity,
        pex_output_file=None if args.no_pex else args.pex_output_file,
        local=args.local,
        serve=args.serve
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import base64
import os.path
from contextlib import contextmanager
from threading import Thread

import colors
import pytest

from pex.cli.testing import run_pex3
from pex.common import safe_rmtree
from pex.compatibility import PY2
from pex.testing import IntegResults, make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterator

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Address(object):
    host = attr.ib()  # type: str
    port = attr.ib()  # type: int


@contextmanager
def serve_authenticated(username, password, find_links):
    expected_authorization = "Basic {}".format(
        base64.b64encode(
            "{username}:{password}".format(username=username, password=password).encode("utf8")
        ).decode("utf-8")
    )

    if PY2:
        from BaseHTTPServer import HTTPServer as HTTPServer
        from SimpleHTTPServer import SimpleHTTPRequestHandler as SimpleHTTPRequestHandler
    else:
        from http.server import HTTPServer as HTTPServer
        from http.server import SimpleHTTPRequestHandler as SimpleHTTPRequestHandler

    class BasicHTTPAuthHandler(SimpleHTTPRequestHandler):
        def do_GET(self):
            authorization = self.headers.get("Authorization")
            if expected_authorization == authorization:
                SimpleHTTPRequestHandler.do_GET(self)
            else:
                self.send_response(401)
                self.send_header("WWW-Authenticate", 'Basic realm="Foo"')
                self.end_headers()

    server = HTTPServer(("", 0), BasicHTTPAuthHandler)
    server_dispatch_thread = Thread(target=server.serve_forever)
    server_dispatch_thread.daemon = True
    cwd = os.getcwd()
    try:
        os.chdir(find_links)
        server_dispatch_thread.start()
        host, port = server.server_address
        yield Address(host=host, port=port)
    finally:
        server.shutdown()
        server_dispatch_thread.join()
        os.chdir(cwd)


@pytest.fixture(scope="module")
def ansicolors_find_links_directory(tmpdir_factory):
    # type: (Any) -> str
    find_links = str(tmpdir_factory.mktemp("find_links"))
    run_pex_command(
        args=[
            "ansicolors==1.1.8",
            "--include-tools",
            "--",
            "repository",
            "extract",
            "--find-links",
            find_links,
        ],
        env=make_env(PEX_TOOLS=1),
    ).assert_success()
    return find_links


@attr.s(frozen=True)
class SecuredLock(object):
    repo_address = attr.ib()  # type: Address
    repo_username = attr.ib()  # type: str
    repo_password = attr.ib()  # type: str
    lock = attr.ib()  # type: str
    pex_root = attr.ib()  # type: str

    @property
    def repo_url(self):
        # type: () -> str
        return "http://{host}:{port}".format(
            host=self.repo_address.host, port=self.repo_address.port
        )

    @property
    def repo_url_with_credentials(self):
        # type: () -> str
        return "http://{username}:{password}@{host}:{port}".format(
            username=self.repo_username,
            password=self.repo_password,
            host=self.repo_address.host,
            port=self.repo_address.port,
        )


@pytest.fixture
def secured_ansicolors_lock(
    ansicolors_find_links_directory,  # type: str
    tmpdir,  # type: Any
):
    # type: (...) -> Iterator[SecuredLock]

    username = "joe"
    password = "bob"
    with serve_authenticated(
        username=username,
        password=password,
        find_links=ansicolors_find_links_directory,
    ) as address:
        lock = os.path.join(str(tmpdir), "lock")
        pex_root = os.path.join(str(tmpdir), "pex_root")
        secured_lock = SecuredLock(
            repo_address=address,
            repo_username=username,
            repo_password=password,
            lock=lock,
            pex_root=pex_root,
        )

        run_pex3(
            "lock",
            "create",
            "--pex-root",
            pex_root,
            "--no-pypi",
            "--find-links",
            secured_lock.repo_url_with_credentials,
            # Since we have no PyPI access, ensure we're using vendored Pip for this test.
            "--pip-version=vendored",
            "ansicolors",
            "--indent",
            "2",
            "-o",
            lock,
        ).assert_success()
        yield secured_lock


def assert_unauthorized(
    secured_ansicolors_lock,  # type: SecuredLock
    result,  # type: IntegResults
):
    # type: (...) -> IntegResults
    result.assert_failure()
    assert (
        "There was 1 error downloading required artifacts:\n"
        "1. ansicolors 1.1.8 from {repo_url}/ansicolors-1.1.8-py2.py3-none-any.whl\n"
        "    ERROR: Could not install requirement ansicolors==1.1.8 from "
        "{repo_url}/ansicolors-1.1.8-py2.py3-none-any.whl because of HTTP error 401 Client Error: "
        "Unauthorized for url: {repo_url}/ansicolors-1.1.8-py2.py3-none-any.whl for URL "
        "{repo_url}/ansicolors-1.1.8-py2.py3-none-any.whl".format(
            repo_url=secured_ansicolors_lock.repo_url
        )
    ) in result.error, result.error

    return result


def test_authenticated_lock_url_issue_1753(
    tmpdir,  # type: Any
    secured_ansicolors_lock,  # type: SecuredLock
):
    # type: (...) -> None

    use_lock_command_unauthenticated = [
        "--pex-root",
        secured_ansicolors_lock.pex_root,
        "--lock",
        secured_ansicolors_lock.lock,
        "--",
        "-c",
        "import colors; print(colors.red('Authenticated'))",
    ]

    def assert_success(result):
        # type: (IntegResults) -> None
        result.assert_success()
        assert colors.red("Authenticated") == result.output.strip()

    # N.B.: Since we created the lock locally, the Pex cache will contain the artifacts needed
    # and no fetches will need to be performed; so, even though we're running without
    # credentials, we should succeed anyhow.
    assert_success(run_pex_command(args=use_lock_command_unauthenticated))

    # But with the Pex caches cleared, fetches should be forced and they should fail without
    # credentials.
    safe_rmtree(secured_ansicolors_lock.pex_root)
    assert_unauthorized(
        secured_ansicolors_lock, run_pex_command(args=use_lock_command_unauthenticated)
    )

    # The find links repo URL without embedded credentials shouldn't help.
    assert_unauthorized(
        secured_ansicolors_lock,
        run_pex_command(
            args=[
                "--find-links",
                secured_ansicolors_lock.repo_url,
            ]
            + use_lock_command_unauthenticated
        ),
    )

    assert_success(
        run_pex_command(
            args=[
                "--find-links",
                secured_ansicolors_lock.repo_url_with_credentials,
            ]
            + use_lock_command_unauthenticated
        )
    )


def test_authenticated_lock_netrc_issue_1753(
    tmpdir,  # type: Any
    secured_ansicolors_lock,  # type: SecuredLock
):
    # type: (...) -> None

    # We don't expect the ambient ~/.netrc, if present, will have the right credentials for an
    # ephemeral port server.
    use_lock_command = [
        "--pex-root",
        secured_ansicolors_lock.pex_root,
        "--lock",
        secured_ansicolors_lock.lock,
        "--",
        "-c",
        "import colors; print(colors.blue('Login Successful'))",
    ]
    safe_rmtree(secured_ansicolors_lock.pex_root)
    assert_unauthorized(secured_ansicolors_lock, run_pex_command(args=use_lock_command))

    # This explicitly controlled ~/.netrc definitely doesn't have the right credentials.
    home = os.path.join(str(tmpdir), "home")
    os.mkdir(home)
    with open(os.path.join(home, ".netrc"), "w") as fp:
        print("machine foo login bar password baz", file=fp)
    assert_unauthorized(
        secured_ansicolors_lock, run_pex_command(args=use_lock_command, env=make_env(HOME=home))
    )

    def assert_authorized(result):
        # type: (IntegResults) -> None
        result.assert_success()
        assert colors.blue("Login Successful") == result.output.strip()

    with open(os.path.join(home, ".netrc"), "a") as fp:
        print(
            "machine {host}:{port} login {username} password {password}".format(
                host=secured_ansicolors_lock.repo_address.host,
                port=secured_ansicolors_lock.repo_address.port,
                username=secured_ansicolors_lock.repo_username,
                password=secured_ansicolors_lock.repo_password,
            ),
            file=fp,
        )
    assert_authorized(run_pex_command(args=use_lock_command, env=make_env(HOME=home)))

    with open(os.path.join(home, ".netrc"), "w") as fp:
        print(
            "machine {url} login {username} password {password}".format(
                url=secured_ansicolors_lock.repo_url,
                username=secured_ansicolors_lock.repo_username,
                password=secured_ansicolors_lock.repo_password,
            ),
            file=fp,
        )
    safe_rmtree(secured_ansicolors_lock.pex_root)
    assert_authorized(run_pex_command(args=use_lock_command, env=make_env(HOME=home)))

    with open(os.path.join(home, ".netrc"), "w") as fp:
        print(
            "default login {username} password {password}".format(
                username=secured_ansicolors_lock.repo_username,
                password=secured_ansicolors_lock.repo_password,
            ),
            file=fp,
        )
    safe_rmtree(secured_ansicolors_lock.pex_root)
    assert_authorized(run_pex_command(args=use_lock_command, env=make_env(HOME=home)))


def test_bad_netrc_issue_1762(
    tmpdir,  # type: Any
    secured_ansicolors_lock,  # type: SecuredLock
):
    # type: (...) -> None

    use_lock_command = [
        "--pex-root",
        secured_ansicolors_lock.pex_root,
        "--lock",
        secured_ansicolors_lock.lock,
        "--",
        "-c",
        "import colors; print(colors.yellow('Welcome'))",
    ]
    safe_rmtree(secured_ansicolors_lock.pex_root)

    home = os.path.join(str(tmpdir), "home")
    os.mkdir(home)
    netrc_path = os.path.join(home, ".netrc")
    with open(netrc_path, "w") as fp:
        print("default login foo password bar", file=fp)
        print("machine foo login bar password baz", file=fp)
        print(
            "machine {host}:{port} login {username} password {password} protocol http".format(
                host=secured_ansicolors_lock.repo_address.host,
                port=secured_ansicolors_lock.repo_address.port,
                username=secured_ansicolors_lock.repo_username,
                password=secured_ansicolors_lock.repo_password,
            ),
            file=fp,
        )

    def assert_netrc_skipped(result):
        # type: (IntegResults) -> None
        assert (
            "Failed to load netrc credentials: bad follower token 'protocol' ({netrc}, line 3)\n"
            "Continuing without netrc credentials.".format(netrc=netrc_path)
        ) in result.error

    result = assert_unauthorized(
        secured_ansicolors_lock,
        run_pex_command(args=use_lock_command, env=make_env(HOME=home), quiet=True),
    )
    assert_netrc_skipped(result)

    result = run_pex_command(
        args=["--find-links", secured_ansicolors_lock.repo_url_with_credentials] + use_lock_command,
        env=make_env(HOME=home),
        quiet=True,
    )
    result.assert_success()
    assert_netrc_skipped(result)
    assert colors.yellow("Welcome") == result.output.strip()



#!/usr/bin/env python3

import os
import subprocess
import sys
from pathlib import Path
from typing import Iterable, Iterator, Sequence

from pex.common import safe_mkdtemp


def find_files_to_check(include: Iterable[str], exclude: Iterable[str] = ()) -> Iterator[str]:
    excluded = frozenset(os.path.normpath(e) for e in exclude)
    for root, dirs, files in os.walk(os.curdir):
        if os.path.realpath(root) == os.path.realpath(os.curdir):
            dirs[:] = list(include)
        else:
            dirs[:] = [
                d for d in dirs if os.path.relpath(os.path.join(root, d), os.curdir) not in excluded
            ]

        for f in files:
            if f.endswith(".py"):
                yield os.path.join(root, f)


def run_mypy(python_version: str, files: Sequence[str], subject: str = "files") -> None:
    print(
        f"Typechecking {len(files)} {subject} using Python "
        f"{'.'.join(map(str, sys.version_info[:3]))} against Python {python_version} ..."
    )
    with (Path(safe_mkdtemp()) / "files.txt").open(mode="w") as fp:
        for f in sorted(files):
            print(f, file=fp)
        fp.close()

        subprocess.run(args=["mypy", "--python-version", python_version, f"@{fp.name}"], check=True)


def main() -> None:
    run_mypy("3.8", files=sorted(find_files_to_check(include=["scripts"])), subject="scripts")

    source_and_tests = sorted(
        find_files_to_check(include=["pex", "tests"], exclude=["pex/vendor/_vendored"])
    )
    for python_version in ("3.11", "3.10", "3.5", "2.7"):
        run_mypy(python_version, files=source_and_tests)


if __name__ == "__main__":
    try:
        main()
    except subprocess.CalledProcessError as e:
        sys.exit(e.returncode)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import hashlib
import json
import os

from pex import hashing
from pex.atomic_directory import FileLockStyle, atomic_directory
from pex.common import safe_rmtree
from pex.pep_503 import ProjectName
from pex.resolve.downloads import get_downloads_dir
from pex.resolve.locked_resolve import Artifact
from pex.result import Error, ResultError, try_
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, Generic

if TYPE_CHECKING:
    from typing import List, Optional, TypeVar, Union

    import attr  # vendor:skip

    from pex.hashing import HintedDigest

    _A = TypeVar("_A", bound=Artifact)
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class DownloadedArtifact(object):
    _METADATA_VERSION = 1

    class LoadError(Exception):
        pass

    @staticmethod
    def metadata_filename(artifact_dir):
        # type: (str) -> str
        return os.path.join(artifact_dir, "metadata.json")

    @classmethod
    def store(
        cls,
        artifact_dir,  # type: str
        filename,  # type: str
        legacy_fingerprint,  # type: hashing.Sha1Fingerprint
        fingerprint,  # type: hashing.Fingerprint
    ):
        # type: (...) -> None

        # We store v0 metadata so that Pex downgrades work. The v0 metadata mechanism had no
        # remove & retry upgrade (downgrade) mechanism like v1+ does.
        with open(os.path.join(artifact_dir, "sha1"), "w") as fp:
            fp.write(legacy_fingerprint)

        with open(cls.metadata_filename(artifact_dir), "w") as fp:
            json.dump(
                dict(
                    algorithm=fingerprint.algorithm,
                    hexdigest=fingerprint,
                    filename=filename,
                    version=cls._METADATA_VERSION,
                ),
                fp,
            )

    @classmethod
    def load(cls, artifact_dir):
        # type: (str) -> DownloadedArtifact
        metadata_filename = cls.metadata_filename(artifact_dir)
        try:
            with open(metadata_filename, "r") as fp:
                try:
                    metadata = json.load(fp)
                except ValueError as e:
                    raise cls.LoadError(
                        "Failed to decode download artifact JSON metadata in {path}: {err}".format(
                            path=metadata_filename, err=e
                        )
                    )
                if not isinstance(metadata, dict) or cls._METADATA_VERSION != metadata.get(
                    "version"
                ):
                    raise cls.LoadError(
                        "Unexpected downloaded artifact metadata object. Expected JSON metadata "
                        "version {version} but found {metadata}".format(
                            version=cls._METADATA_VERSION, metadata=metadata
                        )
                    )
                return DownloadedArtifact(
                    path=os.path.join(artifact_dir, metadata["filename"]),
                    fingerprint=hashing.new_fingerprint(
                        algorithm=str(metadata["algorithm"]), hexdigest=str(metadata["hexdigest"])
                    ),
                )
        except (OSError, IOError) as e:
            raise cls.LoadError(
                "Failed to read a downloaded artifact metadata file at {path}: {err}".format(
                    path=metadata_filename, err=e
                )
            )

    path = attr.ib()  # type: str
    fingerprint = attr.ib()  # type: hashing.Fingerprint


class DownloadManager(Generic["_A"]):
    def __init__(
        self,
        pex_root=None,  # type: Optional[str]
        file_lock_style=FileLockStyle.POSIX,  # type: FileLockStyle.Value
    ):
        # type: (...) -> None
        self._pex_root = pex_root
        self._file_lock_style = file_lock_style

    def store(
        self,
        artifact,  # type: _A
        project_name,  # type: ProjectName
        retry=True,  # type: bool
    ):
        # type: (...) -> DownloadedArtifact

        download_dir = os.path.join(
            get_downloads_dir(pex_root=self._pex_root), artifact.fingerprint.hash
        )
        with atomic_directory(download_dir, lock_style=self._file_lock_style) as atomic_dir:
            if atomic_dir.is_finalized():
                TRACER.log("Using cached artifact at {} for {}".format(download_dir, artifact))
            else:
                legacy_internal_fingerprint = hashing.Sha1()  # Legacy internal
                internal_fingerprint = hashing.Sha256()  # Internal
                digests = [
                    legacy_internal_fingerprint,
                    internal_fingerprint,
                ]  # type: List[HintedDigest]

                # The locking process will have pre-calculated some artifact fingerprints ahead of
                # time; these will be marked as verified and can be trusted.
                if not artifact.verified:
                    # For the rest (E.G.: PyPI URLs with embedded fingerprints), we distrust and
                    # establish our own fingerprint. This will mostly be wasted effort since we
                    # share the same hash algorithm as PyPI currently, but it will serve to upgrade
                    # the fingerprint on other cheeseshops that use a lower-grade hash algorithm (
                    # See: https://peps.python.org/pep-0503/#specification).
                    check = hashlib.new(artifact.fingerprint.algorithm)  # External
                    digests.append(check)

                with TRACER.timed("Downloading {artifact}".format(artifact=artifact)):
                    filename = try_(
                        self.save(
                            artifact=artifact,
                            project_name=project_name,
                            dest_dir=atomic_dir.work_dir,
                            digest=hashing.MultiDigest(digests),
                        )
                    )

                if not artifact.verified:
                    actual_hash = check.hexdigest()
                    if artifact.fingerprint.hash != actual_hash:
                        raise ResultError(
                            Error(
                                "Expected {algorithm} hash of {expected_hash} when downloading "
                                "{project_name} but hashed to {actual_hash}.".format(
                                    algorithm=artifact.fingerprint.algorithm,
                                    expected_hash=artifact.fingerprint.hash,
                                    project_name=project_name,
                                    actual_hash=actual_hash,
                                )
                            )
                        )

                DownloadedArtifact.store(
                    artifact_dir=atomic_dir.work_dir,
                    filename=filename,
                    legacy_fingerprint=legacy_internal_fingerprint.hexdigest(),
                    fingerprint=internal_fingerprint.hexdigest(),
                )
        try:
            return DownloadedArtifact.load(download_dir)
        except DownloadedArtifact.LoadError as e:
            if not retry:
                raise ResultError(Error(str(e)))

            TRACER.log(
                "Found outdated downloaded artifact metadata, upgrading: {err}".format(err=e)
            )
            safe_rmtree(download_dir)
            return self.store(artifact, project_name, retry=False)

    def save(
        self,
        artifact,  # type: _A
        project_name,  # type: ProjectName
        dest_dir,  # type: str
        digest,  # type: HintedDigest
    ):
        # type: (...) -> Union[str, Error]
        """Save and digest the given `artifact` under `dest_dir` and return the saved file name."""
        raise NotImplementedError()



# coding=utf-8
# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import platform
import subprocess
import sys

from pex.common import safe_rmtree
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_unicode_script_shebang_rewrite_docutils(tmpdir):
    # type: (Any) -> None

    # The docutils distribution contains many scripts, one of which, rst2html5.py, contains
    # non-ascii unicode characters which could trip up script shebang re-writing in environments
    # without a default encoding accepting those characters.

    pex_root = os.path.join(str(tmpdir), "pex_root")
    env = make_env(LANG=None, PEX_ROOT=pex_root)

    docutils_pex = os.path.join(str(tmpdir), "docutils.pex")

    run_pex_command(
        args=[
            "docutils==0.17.1",
            "-c",
            "rst2html5.py",
            "-o",
            docutils_pex,
            "--venv",
        ],
        env=env,
    ).assert_success()

    safe_rmtree(pex_root)
    output = subprocess.check_output(args=[docutils_pex, "-V"], env=env)
    assert "rst2html5.py (Docutils 0.17.1 [release], Python {version}, on {platform})\n".format(
        version=platform.python_version(), platform=sys.platform
    ) == output.decode("utf-8")



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import filecmp
import os.path
import shutil
import subprocess
import sys
import tempfile
from textwrap import dedent

import colors
import pytest

from pex.cli.testing import run_pex3
from pex.common import safe_open
from pex.compatibility import urlparse
from pex.resolve.locked_resolve import VCSArtifact
from pex.resolve.lockfile import json_codec
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class TestTool(object):
    __test__ = False

    tmpdir = attr.ib()  # type: str

    @property
    def pex_root(self):
        # type: () -> str
        return os.path.join(self.tmpdir, "pex_root")

    def create_lock(self, *args):
        # type: (*str) -> str
        lock = tempfile.mktemp(prefix="lock.", dir=self.tmpdir)
        run_pex3("lock", "create", "--pex-root", self.pex_root, "-o", lock, *args).assert_success()
        return lock

    def create_pex(
        self,
        lock,  # type: str
        *args  # type: str
    ):
        # type: (...) -> str
        pex = tempfile.mktemp(suffix=".pex", dir=self.tmpdir)
        run_pex_command(
            args=[
                "--pex-root",
                self.pex_root,
                "--runtime-pex-root",
                self.pex_root,
                "--lock",
                lock,
                "-o",
                pex,
            ]
            + list(args)
        ).assert_success()
        return pex

    def create_locked_pex(self, *lock_args):
        # type: (*str) -> str
        return self.create_pex(self.create_lock(*lock_args))


@pytest.fixture
def test_tool(tmpdir):
    # type: (Any) -> TestTool
    return TestTool(str(tmpdir))


def test_vcs_direct_reference(test_tool):
    # type: (TestTool) -> None

    pex = test_tool.create_locked_pex(
        "ansicolors @ git+https://github.com/jonathaneunice/colors.git@c965f5b9"
    )

    assert (
        colors.cyan("Miles Davis")
        == subprocess.check_output(
            args=[pex, "-c", "import colors; print(colors.cyan('Miles Davis'))"]
        )
        .decode("utf-8")
        .strip()
    )


def test_vcs_pip_proprietary(test_tool):
    # type: (TestTool) -> None

    pex = test_tool.create_locked_pex(
        "git+https://github.com/jonathaneunice/colors.git@c965f5b9#egg=ansicolors"
    )

    assert (
        colors.magenta("Alecia Beth Moore")
        == subprocess.check_output(
            args=[pex, "-c", "import colors; print(colors.magenta('Alecia Beth Moore'))"]
        )
        .decode("utf-8")
        .strip()
    )


def test_vcs_equivalence(test_tool):
    # type: (TestTool) -> None

    lock1 = test_tool.create_lock(
        "ansicolors @ git+https://github.com/jonathaneunice/colors.git@c965f5b9"
    )
    lock2 = test_tool.create_lock(
        "git+https://github.com/jonathaneunice/colors.git@c965f5b9#egg=ansicolors"
    )

    assert lock1 != lock2, "Expected two different lock files."

    def extract_single_vcs_artifact(lock):
        # type: (str) -> VCSArtifact
        lock_file = json_codec.load(lock)

        assert 1 == len(lock_file.locked_resolves)
        locked_resolve = lock_file.locked_resolves[0]

        assert 1 == len(locked_resolve.locked_requirements)
        locked_requirement = locked_resolve.locked_requirements[0]

        assert 0 == len(locked_requirement.additional_artifacts)
        assert isinstance(locked_requirement.artifact, VCSArtifact)
        return locked_requirement.artifact

    vcs_artifact1 = extract_single_vcs_artifact(lock1)
    vcs_artifact2 = extract_single_vcs_artifact(lock2)

    assert vcs_artifact1.fingerprint == vcs_artifact2.fingerprint, (
        "We expect locking using a direct reference requirement or a Pip proprietary VCS "
        "requirement for the same VCS revision will produce the same locked VCS archive"
    )
    assert vcs_artifact1.url != vcs_artifact2.url, "Expected two different lock URLs."
    assert "" == urlparse.urlparse(vcs_artifact1.url).fragment
    assert {"egg": ["ansicolors"]} == urlparse.parse_qs(
        urlparse.urlparse(vcs_artifact2.url).fragment
    )


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="The library under test uses f-strings.")
def test_subdir(test_tool):
    # type: (TestTool) -> None

    # This is a trick you cannot do with a direct reference VCS URL.
    pex = test_tool.create_locked_pex(
        "git+https://github.com/SerialDev/sdev_py_utils.git@bd4d36a0"
        "#egg=sdev_logging_utils&subdirectory=sdev_logging_utils"
    )
    assert (
        subprocess.check_output(
            args=[pex, "-c", "import sdev_logging_utils; print(sdev_logging_utils.__file__)"]
        )
        .decode("utf-8")
        .startswith(os.path.join(test_tool.pex_root, "installed_wheels"))
    )


def test_vcs_fingerprint_stability(test_tool):
    # type: (TestTool) -> None

    lock1 = test_tool.create_lock(
        "git+https://github.com/VaasuDevanS/cowsay-python@v3.0#egg=cowsay"
    )
    shutil.rmtree(test_tool.pex_root)
    lock2 = test_tool.create_lock(
        "git+https://github.com/VaasuDevanS/cowsay-python@v3.0#egg=cowsay"
    )

    assert lock1 != lock2, "Expected two different lock files."
    assert filecmp.cmp(
        lock1, lock2, shallow=False
    ), "Expected the same lockfile contents; i.e.: a stable VCS archive hash."


def test_vcs_transitive(
    tmpdir,  # type: Any
    test_tool,  # type: TestTool
):
    # type: (...) -> None

    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "poetry.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import colors


                def third_worst():
                    print(colors.green("Prostetnic Vogon Jeltz"))
                """
            )
        )
    with safe_open(os.path.join(src, "setup.cfg"), "w") as fp:
        fp.write(
            dedent(
                """\
                [metadata]
                name = poetry
                version = 0.0.1

                [options]
                py_modules =
                    poetry

                install_requires =
                    ansicolors @ git+https://github.com/jonathaneunice/colors.git@c965f5b9

                [options.entry_points]
                console_scripts =
                    recite = poetry:third_worst
                """
            )
        )
    with safe_open(os.path.join(src, "setup.py"), "w") as fp:
        fp.write("from setuptools import setup; setup()")

    subprocess.check_call(args=["git", "init", src])
    subprocess.check_call(args=["git", "config", "user.email", "forty@two.com"], cwd=src)
    subprocess.check_call(args=["git", "config", "user.name", "Douglas Adams"], cwd=src)
    subprocess.check_call(args=["git", "checkout", "-b", "Golgafrincham"], cwd=src)
    subprocess.check_call(args=["git", "add", "."], cwd=src)
    subprocess.check_call(args=["git", "commit", "-m", "Only commit."], cwd=src)

    lock = test_tool.create_lock("git+file://{src}#egg=poetry".format(src=src))
    pex = test_tool.create_pex(lock, "-c", "recite")
    assert (
        colors.green("Prostetnic Vogon Jeltz")
        == subprocess.check_output(args=[pex]).decode("utf-8").strip()
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import json
import os.path
import subprocess

from pex import dist_metadata
from pex.dist_metadata import ProjectNameAndVersion
from pex.orderedset import OrderedSet
from pex.pex_info import PexInfo
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_duplicate_requirements_issues_1550(tmpdir):
    # type: (Any) -> None

    pex_file = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "PyJWT",
            "PyJWT==1.7.1",
            "--resolver-version",
            "pip-2020-resolver",
            "-o",
            pex_file,
        ]
    ).assert_success()

    subprocess.check_call(args=[pex_file, "-c", "import jwt"])
    pex_info = PexInfo.from_pex(pex_file)
    assert 1 == len(pex_info.distributions)
    assert ProjectNameAndVersion("PyJWT", "1.7.1") == dist_metadata.project_name_and_version(
        next(iter(pex_info.distributions.keys()))
    ), "ex_info.distributions: {}".format(pex_info.distributions)
    assert OrderedSet(("PyJWT", "PyJWT==1.7.1")) == pex_info.requirements



# Copyright 2016 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import os

from pex.compatibility import PY2, string
from pex.tracer import TRACER

if os.name == "posix" and PY2:
    try:
        # Use the subprocess backports if they're available for improved robustness.
        import subprocess32 as subprocess  # type: ignore[import]
    except ImportError:
        TRACER.log(
            "Please build pex with the subprocess32 module for more reliable requirement "
            "installation and interpreter execution."
        )
        import subprocess  # type: ignore[no-redef]
else:
    import subprocess  # type: ignore[no-redef]


class Executor(object):
    """Handles execution of subprocesses in a structured way."""

    class ExecutionError(Exception):
        """Indicates failure to execute."""

        def __init__(self, msg, cmd, exc=None):
            super(Executor.ExecutionError, self).__init__(  # noqa
                "%s while trying to execute `%s`" % (msg, cmd)
            )
            self.executable = cmd.split()[0] if isinstance(cmd, string) else cmd[0]
            self.cmd = cmd
            self.exc = exc

    class NonZeroExit(ExecutionError):
        """Indicates a non-zero exit code."""

        def __init__(self, cmd, exit_code, stdout, stderr):
            super(Executor.NonZeroExit, self).__init__(  # noqa
                "received exit code %s during execution of `%s`" % (exit_code, cmd), cmd
            )
            self.exit_code = exit_code
            self.stdout = stdout
            self.stderr = stderr

    class ExecutableNotFound(ExecutionError):
        """Indicates the executable was not found while attempting to execute."""

        def __init__(self, cmd, exc):
            super(Executor.ExecutableNotFound, self).__init__(  # noqa
                "caught %r while trying to execute `%s`" % (exc, cmd), cmd
            )
            self.exc = exc

    @classmethod
    def open_process(cls, cmd, **kwargs):
        """Opens a process object via subprocess.Popen().

        :param string|list cmd: A list or string representing the command to run.
        :param **kwargs: Additional kwargs to pass through to subprocess.Popen.
        :return: A `subprocess.Popen` object.
        :raises: `Executor.ExecutableNotFound` when the executable requested to run does not exist.
        """
        assert len(cmd) > 0, "cannot execute an empty command!"

        try:
            return subprocess.Popen(cmd, **kwargs)
        except (IOError, OSError) as e:
            if e.errno == errno.ENOENT:
                raise cls.ExecutableNotFound(cmd, e)
            else:
                raise cls.ExecutionError(repr(e), cmd, e)

    @classmethod
    def execute(cls, cmd, stdin_payload=None, **kwargs):
        """Execute a command via subprocess.Popen and returns the stdio.

        :param string|list cmd: A list or string representing the command to run.
        :param string stdin_payload: A string representing the stdin payload, if any, to send.
        :param **kwargs: Additional kwargs to pass through to subprocess.Popen.
        :return: A tuple of strings representing (stdout, stderr), pre-decoded for utf-8.
        :raises: `Executor.ExecutableNotFound` when the executable requested to run does not exist.
                 `Executor.NonZeroExit` when the execution fails with a non-zero exit code.
        """
        process = cls.open_process(
            cmd=cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs
        )
        stdout_raw, stderr_raw = process.communicate(input=stdin_payload)
        # N.B. In cases where `stdout` or `stderr` is passed as parameters, these can be None.
        stdout = stdout_raw.decode("utf-8") if stdout_raw is not None else stdout_raw
        stderr = stderr_raw.decode("utf-8") if stderr_raw is not None else stderr_raw

        if process.returncode != 0:
            raise cls.NonZeroExit(cmd, process.returncode, stdout, stderr)

        return stdout, stderr
